/*
* JBoss, Home of Professional Open Source
* Copyright 2005, JBoss Inc., and individual contributors as indicated
* by the @authors tag. See the copyright.txt in the distribution for a
* full listing of individual contributors.
*
* This is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2.1 of
* the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free
* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
* 02110-1301 USA, or see the FSF site: http://www.fsf.org.
*/

package org.jboss.remoting;

import java.util.HashMap;
import java.util.Map;
import org.jboss.logging.Logger;
import org.jboss.remoting.loading.ClassByteClassLoader;
import org.jboss.remoting.marshal.MarshallLoaderFactory;
import org.jboss.remoting.serialization.SerializationStreamFactory;

/**
 * AbstractInvoker is an abstract handler part that contains common methods between both
 * client and server.
 *
 * @author <a href="mailto:jhaynie@vocalocity.net">Jeff Haynie</a>
 * @author <a href="mailto:telrod@e2technologies.net">Tom Elrod</a>
 * @version $Revision: 1.8 $
 */
public abstract class AbstractInvoker implements Invoker
{
   protected final Logger log = Logger.getLogger(getClass());
   protected ClassByteClassLoader classbyteloader;
   protected InvokerLocator locator;
   protected Map localServerLocators = new HashMap();
   protected String serializationType;
   protected Map configuration = new HashMap();

   public AbstractInvoker(InvokerLocator locator)
   {
      this(locator, null);
   }
   
   public AbstractInvoker(InvokerLocator locator, Map configuration)
   {
      this.classbyteloader = new ClassByteClassLoader(this.getClass().getClassLoader());
      this.locator = locator;
      
      if (configuration != null)
         this.configuration.putAll(configuration);
      
      if (locator.getParameters() != null)
         this.configuration.putAll(locator.getParameters());

      try
      {
         InvokerLocator loaderLocator = MarshallLoaderFactory.convertLocator(locator);
         if(loaderLocator != null)
         {
            classbyteloader.setClientInvoker(new Client(loaderLocator));
         }
      }
      catch(Exception e)
      {
         log.error("Could not create remote class loading for invoker.", e);
      }

      if(locator == null || locator.getParameters() == null)
      {
         this.setSerializationType(SerializationStreamFactory.JAVA);
      }
      else
      {
         this.setSerializationType(locator.findSerializationType());
      }
   }

   /**
    * return the locator this Invoker represents
    *
    * @return
    */
   public InvokerLocator getLocator()
   {
      return locator;
   }

   /**
    * Sets the callback server locator for the specified callback listener id
    *
    * @param locator
    */
   public void addClientLocator(String listenerId, InvokerLocator locator)
   {
      Object obj = localServerLocators.put(listenerId, locator);
      if(obj != null)
      {
         throw new RuntimeException("InvokerLocator already exists for listener id " + listenerId + ".  " +
                                    "Remove this listener before adding again.");
      }
   }

   /**
    * Gets the client locator.  This locator will be used by the server side
    * to make callbacks to the handler for this locator.
    *
    * @return
    */
   public InvokerLocator getClientLocator(String listenerId)
   {
      return (InvokerLocator) localServerLocators.get(listenerId);
   }

   /**
    * set the classloader to use internally
    *
    * @param classloader
    */
   public synchronized void setClassLoader(ClassLoader classloader)
   {
      this.classbyteloader = new ClassByteClassLoader(classloader);
   }

   public synchronized ClassLoader getClassLoader()
   {
      return classbyteloader;
   }

   public String getSerializationType()
   {
      return serializationType;
   }

   public void setSerializationType(String serializationType)
   {
      this.serializationType = serializationType;
   }
}
