/*
* JBoss, Home of Professional Open Source
* Copyright 2005, JBoss Inc., and individual contributors as indicated
* by the @authors tag. See the copyright.txt in the distribution for a
* full listing of individual contributors.
*
* This is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2.1 of
* the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free
* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
* 02110-1301 USA, or see the FSF site: http://www.fsf.org.
*/

package org.jboss.remoting;

import java.util.ArrayList;
import java.util.List;
import java.util.ListIterator;
import java.util.Timer;
import java.util.TimerTask;
import org.jboss.remoting.transport.ClientInvoker;

/**
 * @author <a href="mailto:tom.elrod@jboss.com">Tom Elrod</a>
 */
public class ConnectionValidator extends TimerTask
{
   private List listeners = new ArrayList();
   private Client client = null;

   private Timer validatorTimer = null;
   private long pingPeriod = DEFAULT_PING_PERIOD;

   /**
    * Default ping period.  Value is 2 seconds.
    */
   public static final long DEFAULT_PING_PERIOD = 2000;

   public ConnectionValidator(Client client)
   {
      this.client = client;
   }

   public ConnectionValidator(Client client, int pingPeriod)
   {
      this.client = client;
      this.pingPeriod = pingPeriod;
   }

   private void start()
   {
      validatorTimer = new Timer(false);
      validatorTimer.schedule(this, pingPeriod, pingPeriod);
   }

   private void stop()
   {
      this.cancel();
      validatorTimer.cancel();
      validatorTimer = null;
   }


   public void addConnectionListener(ConnectionListener listener)
   {
      if(listener != null)
      {
         synchronized(listener)
         {
            if(listeners.size() == 0)
            {
               start();
            }
            listeners.add(listener);
         }
      }
   }

   public boolean removeConnectionListener(ConnectionListener listener)
   {
      boolean isRemoved = false;
      if(listener != null)
      {
         synchronized(listeners)
         {
            isRemoved = listeners.remove(listener);
            if(listeners.size() == 0)
            {
               stop();
            }
         }
      }
      return isRemoved;
   }

   /**
    * The action to be performed by this timer task.
    */
   public void run()
   {
      try
      {
         boolean isValid = checkConnection(client.getInvoker());
         if(!isValid)
         {
            notifyListeners(new Exception("Could not connect to server."));
         }
      }
      catch(Throwable thr)
      {
         notifyListeners(thr);
      }
   }

   private void notifyListeners(Throwable thr)
   {
      final Throwable t = thr;
      synchronized(listeners)
      {
         ListIterator itr = listeners.listIterator();
         while(itr.hasNext())
         {
            final ConnectionListener listener = (ConnectionListener) itr.next();
            new Thread()
            {
               public void run()
               {
                  listener.handleConnectionException(t, client);
               }
            }.start();
         }
      }
   }

   public static boolean checkConnection(ClientInvoker clientInvoker) throws Throwable
   {
      Object o = clientInvoker.invoke(new InvocationRequest(ConnectionValidator.class.getName(),
                                                            Subsystem.SELF, "$PING$", null, null, null));
      InvocationResponse resp = (InvocationResponse) o;
      Boolean b = (Boolean) resp.getResult();
      return b.booleanValue();
   }


}