/*
* JBoss, Home of Professional Open Source
* Copyright 2005, JBoss Inc., and individual contributors as indicated
* by the @authors tag. See the copyright.txt in the distribution for a
* full listing of individual contributors.
*
* This is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2.1 of
* the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free
* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
* 02110-1301 USA, or see the FSF site: http://www.fsf.org.
*/
package org.jboss.remoting;

import java.util.Map;
import org.jboss.util.TimerQueue;
import org.jboss.util.TimerTask;

/**
 * This class is used on the remoting server to maintain lease information
 * for remoting clients.  Will generate callback to ConnectionListener interface
 * if determined that client no longer available.
 *
 * @author <a href="mailto:tom.elrod@jboss.com">Tom Elrod</a>
 */
public class Lease
{
   /**
    * This value (defaults to 5 seconds) represents how much of a
    * delay will be built into starting the lease timer.  This is
    * done since need to give the client some extra buffer time to
    * start sending lease pings.  Otherwise, client might get the
    * lease time and start pinging before the lease time runs out.
    */
   public static long TIMER_DELAY = 5000;

   private ConnectionNotifier notifier = null;
   private String clientSessionId = null;
   private long leasePeriod = -1;
   private String locatorURL = null;
   private Map requestPayload = null;
   private TimerQueue leaseTimer = null;
   private LeaseTimerTask leaseTimerTask = null;
   private long leaseWindow = -1;
   private long pingStart = -1;
   private Map clientLeases = null;

   private boolean leaseUpdated = false;

   public Lease(String clientSessionId, long leasePeriod, String locatorurl, Map requestPayload,
                ConnectionNotifier notifier, Map clientLeases)
   {
      this.clientSessionId = clientSessionId;
      this.leasePeriod = leasePeriod;
      this.notifier = notifier;
      this.locatorURL = locatorurl;
      this.requestPayload = requestPayload;
      this.leaseWindow = leasePeriod;
      this.clientLeases = clientLeases;
      leaseTimer = new TimerQueue("Lease-" + clientSessionId);
   }


   public void startLease()
   {
      leaseTimer.start();
      leaseTimerTask = new LeaseTimerTask();
      leaseTimer.schedule(leaseTimerTask, leasePeriod + TIMER_DELAY);
      pingStart = System.currentTimeMillis();
   }

   public void updateLease(long leasePeriod)
   {
      leaseUpdated = true;
      if(leasePeriod != this.leasePeriod)
      {
         this.leasePeriod = leasePeriod;
         this.leaseWindow = leasePeriod;
         stopLease();
         startLease();
      }
      else
      {
         long pingDuration = System.currentTimeMillis() - pingStart;
         if(pingDuration > (leaseWindow / 2))
         {
            leaseWindow = pingDuration * 2;
         }
         leaseTimer.schedule(leaseTimerTask, leaseWindow);
         pingStart = System.currentTimeMillis();
      }

         }

   public void terminateLease(long leasePeriod)
   {
      stopLease();
      notifier.connectionTerminated(locatorURL, clientSessionId, requestPayload);
   }

   private void stopLease()
   {
      leaseTimerTask.running = false;
      leaseTimer.stop();
   }

   private class LeaseTimerTask extends TimerTask
   {

      protected boolean running = true;

      /**
       * The action to be performed by this timer task.
       */
      public void execute()
      {
         if(running)
         {
            if(leaseUpdated)
            {
               leaseUpdated = false;
               leaseTimer.schedule(this, leaseWindow);
            }
            else
            {
               stopLease();
               notifier.connectionLost(locatorURL, clientSessionId, requestPayload);
               if(clientLeases != null)
               {
                  clientLeases.remove(clientSessionId);
            }
         }
      }
   }
}}