/*
* JBoss, Home of Professional Open Source
* Copyright 2005, JBoss Inc., and individual contributors as indicated
* by the @authors tag. See the copyright.txt in the distribution for a
* full listing of individual contributors.
*
* This is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2.1 of
* the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free
* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
* 02110-1301 USA, or see the FSF site: http://www.fsf.org.
*/
package org.jboss.remoting.security;

import java.io.IOException;
import java.security.KeyManagementException;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.UnrecoverableKeyException;
import java.security.cert.CertificateException;
import javax.net.ServerSocketFactory;
import javax.net.SocketFactory;

/**
 * @author <a href="mailto:telrod@e2technologies.net">Tom Elrod</a>
 */
public interface SSLSocketBuilderMBean
{
   /**
    * create the service, do expensive operations etc
    */
   void create() throws Exception;

   /**
    * start the service, create is already called
    */
   void start() throws Exception;

   /**
    * stop the service
    */
   void stop();

   /**
    * destroy the service, tear down
    */
   void destroy();

   /**
    * Will indicate if should use the SSLServerSocketFactory.getDefault() for getting the ServerSocketFactory
    * to use (when calling createSSLServerSocketFactory()).
    * If true, will allow for setting key store location (via javax.net.ssl.keyStore system property) and
    * setting of the key store password (via javax.net.ssl.keyStorePassword system property) and no other
    * configuration is needed (none of the other setters will need to be called and are in fact ignored).  If set to
    * false, will allow the custom setting of secure socket protocol, key management algorithm, key store type,
    * key store url, key store password, and key password.<p/>
    * The default value is true.<p/>
    * <b>NOTE: If this is not explicitly set to false, no customizations can be made and the default implementation
    * provided by the JVM vendor being used will be executed.</b>
    *
    * @param shouldUse
    */
   void setUseSSLServerSocketFactory(boolean shouldUse);

   /**
    * Return whether SSLServerSocketFactory.getDefault() will be used or not.  See setUseSSLServerSocketFactory() for more
    * information on what this means.
    *
    * @return
    */
   boolean getUseSSLServerSocketFactory();

   /**
    * Will indicate if should use the SSLSocketFactory.getDefault() for getting the SocketFactory
    * to use (when calling createSSLSocketFactory()).
    * If true, will allow for setting trust store location (via Djavax.net.ssl.trustStore system property) and no other
    * configuration is needed (none of the other setters will need to be called and are in fact ignored).  If set to
    * false, will allow the custom setting of secure socket protocol, key management algorithm, key store type,
    * ant trust store url.<p/>
    * The default value is true.<p/>
    * <b>NOTE: If this is not explicitly set to false, no customizations can be made and the default implementation
    * provided by the JVM vendor being used will be executed.</b>
    *
    * @param shouldUse
    */
   void setUseSSLSocketFactory(boolean shouldUse);

   /**
    * Return whether SSLSocketFactory.getDefault() will be used or not.  See setUseSSLSocketFactory() for more
    * information on what this means.
    *
    * @return
    */
   boolean getUseSSLSocketFactory();

   /**
    * The protocol for the SSLContext.  Some acceptable values are TLS, SSL, and SSLv3.
    * Defaults to DEFAULT_SECURE_SOCKET_PROTOCOL.
    */
   String getSecureSocketProtocol();

   /**
    * The protocol for the SSLContext.  Some acceptable values are TLS, SSL, and SSLv3.
    * Defaults to DEFAULT_SECURE_SOCKET_PROTOCOL.
    */
   void setSecureSocketProtocol(String secureSocketProtocol);

   /**
    * The algorithm for the key manager factory.
    * Defaults to DEFAULT_KEY_MANAGEMENT_ALGORITHM.
    */
   String getKeyManagementAlgorithm();

   /**
    * The algorithm for the key manager factory.
    * Defaults to DEFAULT_KEY_MANAGEMENT_ALGORITHM.
    */
   void setKeyManagementAlgorithm(String keyManagementAlgorithm);

   /**
    * The type to be used for the key store.
    * Defaults to DEFAULT_KEY_STORE_TYPE.  Some acceptable values are JKS (Java Keystore - Sun's keystore format),
    * JCEKS (Java Cryptography Extension keystore - More secure version of JKS), and
    * PKCS12 (Public-Key Cryptography Standards #12 keystore - RSA's Personal Information Exchange Syntax Standard).
    * These are not case sensitive.
    */
   String getKeyStoreType();

   /**
    * The type to be used for the key store.
    * Defaults to DEFAULT_KEY_STORE_TYPE.  Some acceptable values are JKS (Java Keystore - Sun's keystore format),
    * JCEKS (Java Cryptography Extension keystore - More secure version of JKS), and
    * PKCS12 (Public-Key Cryptography Standards #12 keystore - RSA's Personal Information Exchange Syntax Standard).
    * These are not case sensitive.
    */
   void setKeyStoreType(String keyStoreType);

   /**
    * Sets the password to use for the key store.  This only needs to be set if setUseSSLServerSocketFactory() is set
    * to false (otherwise will be ignored).  The value passed will also be used for the key password if it is not
    * explicitly set.
    *
    * @param passphrase
    */
   void setKeyStorePassword(String passphrase);

   /**
    * Sets the password to use for the keys within the key store.  This only needs to be set if setUseSSLServerSocketFactory()
    * is set to false (otherwise will be ignored).  If this value is not set, but the key store password is, it will use
    * that value for the key password.
    *
    * @param passphrase
    */
   void setKeyPassword(String passphrase);

   /**
    * Will create a SSLServerSocketFactory.  If the useSSLServerSocketFactory property is set to true (which is the default),
    * it will use SSLServerSocketFactory.getDefault() to get the server socket factory.  Otherwise, if property is false,
    * will use all the other custom properties that have been set to create a custom server socket factory.
    *
    * @return
    * @throws java.io.IOException
    * @throws java.security.NoSuchAlgorithmException
    *
    * @throws java.security.KeyStoreException
    *
    * @throws java.security.cert.CertificateException
    *
    * @throws java.security.UnrecoverableKeyException
    *
    * @throws java.security.KeyManagementException
    *
    */
   ServerSocketFactory createSSLServerSocketFactory()
         throws IOException, NoSuchAlgorithmException, KeyStoreException,
                CertificateException, UnrecoverableKeyException, KeyManagementException;

   /**
    * Will create a SSLSocketFactory.  If the useSSLSocketFactory property is set to true (which is the default),
    * it will use SSLSocketFactory.getDefault() to get the socket factory.  Otherwise, if property is false,
    * will use all the other custom properties that have been set to create a custom server socket factory.
    *
    * @return
    * @throws java.io.IOException
    * @throws java.security.NoSuchAlgorithmException
    *
    * @throws java.security.KeyStoreException
    *
    * @throws java.security.cert.CertificateException
    *
    * @throws java.security.KeyManagementException
    *
    */
   SocketFactory createSSLSocketFactory()
         throws IOException, NoSuchAlgorithmException, KeyStoreException,
                CertificateException, KeyManagementException;

   /**
    * This is the url string to the key store to use.  If UseSSLServerSocketFactory is true, this will be ignored
    * and will use the value set by the javax.net.ssl.keyStore system property.  Otherwise, if UseSSLServerSocketFactory
    * is false, this must be set.
    *
    * @param storeURL
    * @throws java.io.IOException
    */
   void setKeyStoreURL(String storeURL) throws IOException;

   /**
    * This is the url string to the trust store to use.  If UseSSLSocketFactory is true, this will be ignored
    * and will use the value set by the javax.net.ssl.trustStore system property.  Otherwise, if UseSSLSocketFactory
    * is false, this must be set.
    *
    * @param storeURL
    * @throws java.io.IOException
    */
   void setTrustStoreURL(String storeURL) throws IOException;
}
