/*
* JBoss, Home of Professional Open Source
* Copyright 2005, JBoss Inc., and individual contributors as indicated
* by the @authors tag. See the copyright.txt in the distribution for a
* full listing of individual contributors.
*
* This is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2.1 of
* the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free
* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
* 02110-1301 USA, or see the FSF site: http://www.fsf.org.
*/

package org.jboss.remoting.transport.http;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.HttpURLConnection;
import java.net.URL;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;
import org.jboss.logging.Logger;
import org.jboss.remoting.CannotConnectException;
import org.jboss.remoting.ConnectionFailedException;
import org.jboss.remoting.InvokerLocator;
import org.jboss.remoting.RemoteClientInvoker;
import org.jboss.remoting.marshal.Marshaller;
import org.jboss.remoting.marshal.UnMarshaller;
import org.jboss.remoting.marshal.http.HTTPMarshaller;
import org.jboss.remoting.transport.web.WebUtil;
import org.jboss.util.Base64;

/**
 * HTTP client invoker.  Used for making http requests on http/servlet invoker.
 *
 * @author <a href="mailto:tom@jboss.org">Tom Elrod</a>
 */
public class HTTPClientInvoker extends RemoteClientInvoker
{

   protected final Logger log = Logger.getLogger(getClass());

   public HTTPClientInvoker(InvokerLocator locator)
   {
      super(locator);
   }

   public HTTPClientInvoker(InvokerLocator locator, Map configuration)
   {
      super(locator, configuration);
   }

   /**
    * @param sessionId
    * @param invocation
    * @param marshaller
    * @return
    * @throws java.io.IOException
    * @throws org.jboss.remoting.ConnectionFailedException
    *
    */
   protected Object transport(String sessionId, Object invocation, Map metadata,
                              Marshaller marshaller, UnMarshaller unmarshaller)
         throws IOException, ConnectionFailedException
   {
      String targetURL = getLocator().getOriginalURI();
      Object httpResponse = useHttpURLConnection(targetURL, invocation, metadata, marshaller, unmarshaller);

      return httpResponse;
   }

   private Object useHttpURLConnection(String url, Object invocation, Map metadata,
                                       Marshaller marshaller, UnMarshaller unmarshaller)
   {
      Object result = null;
      try
      {
         // need to check the url and make sure it compatible protocol
         String validatedUrl = validateURL(url);

         if(metadata == null)
         {
            metadata = new HashMap();
         }

         HttpURLConnection conn = createURLConnection(validatedUrl, metadata);

         setTimeout(configuration, conn);

         // check to see if basic auth required
         String basicAuth = getBasicAuth(metadata);
         if(basicAuth != null)
         {
            conn.setRequestProperty("Authorization", basicAuth);
         }

         // Get the request method type
         boolean sendingData = true;
         String type = "POST";
         if(metadata != null)
         {
            type = (String) metadata.get("TYPE");
            if(type != null)
            {
               if((!type.equals("POST") && !type.equals("PUT")))
               {
                  sendingData = false;
               }
            }
            else
            {
               type = "POST";
            }
         }
         else // need to check for content type and set metadata
         {
            metadata = new HashMap();
            Map header = new HashMap();
            header.put(HTTPMetadataConstants.CONTENTTYPE, WebUtil.getContentType(invocation));
            metadata.put("HEADER", header);
         }
         // Set request headers
         Map header = (Map) metadata.get("HEADER");
         if(header != null)
         {
            Set keys = header.keySet();
            Iterator itr = keys.iterator();
            while(itr.hasNext())
            {
               String key = (String) itr.next();
               String value = (String) header.get(key);
               log.debug("Setting request header with " + key + " : " + value);
               conn.setRequestProperty(key, value);
            }
         }
         else
         {
            conn.setRequestProperty(HTTPMetadataConstants.CONTENTTYPE, WebUtil.getContentType(invocation));
         }


         if(sendingData)
         {
            //POST or PUT
            conn.setDoOutput(true);
            conn.setDoInput(true);
            conn.setRequestMethod(type);

            OutputStream stream = conn.getOutputStream();
            marshaller.write(invocation, stream);
            int responseCode = conn.getResponseCode();
            InputStream is = (responseCode < 400) ? conn.getInputStream() : conn.getErrorStream();
            Map headers = conn.getHeaderFields();
            if(metadata == null)
            {
               metadata = new HashMap();
            }
            metadata.putAll(headers);
            metadata.put(HTTPMetadataConstants.RESPONSE_CODE_MESSAGE, conn.getResponseMessage());
            metadata.put(HTTPMetadataConstants.RESPONSE_CODE, new Integer(responseCode));
            result = unmarshaller.read(is, metadata);
         }
         else
         {
            conn.setDoOutput(false);
            conn.setDoInput(true);
            conn.setRequestMethod(type);

            conn.connect();

            InputStream is = (conn.getResponseCode() < 400) ? conn.getInputStream() : conn.getErrorStream();
            Map headers = conn.getHeaderFields();
            result = unmarshaller.read(is, headers);
            if(metadata == null)
            {
               metadata = new HashMap();
            }
            metadata.putAll(headers);
            metadata.put(HTTPMetadataConstants.RESPONSE_CODE_MESSAGE, conn.getResponseMessage());
            int responseCode = conn.getResponseCode();
            metadata.put(HTTPMetadataConstants.RESPONSE_CODE, new Integer(responseCode));
         }
      }
      catch(Exception e)
      {
         log.debug("Error invoking http client invoker.", e);
         throw new CannotConnectException("Can not connect http client invoker.", e);
      }

      return result;
   }

   private void setTimeout(Map metadata, HttpURLConnection conn)
   {
      String timeoutValue = (String)metadata.get("timeout");
      if(timeoutValue != null && timeoutValue.length() > 0)
      {
         try
         {
            int timeout = Integer.parseInt(timeoutValue);
            /**
             * Since URLConnection in jdk 1.4 does NOT have a setConnectTimeout() method and
             * the one in jdk 1.5 does, will have to use reflection to see if it exists before
             * trying to set it.
             */
            try
            {
               Method setTimeoutMethod = conn.getClass().getMethod("setConnectTimeout", new Class[] {int.class});
               setTimeoutMethod.invoke(conn, new Object[] {new Integer(timeout)});
               setTimeoutMethod = conn.getClass().getMethod("setReadTimeout", new Class[] {int.class});
               setTimeoutMethod.invoke(conn, new Object[] {new Integer(timeout)});
            }
            catch(NoSuchMethodException e)
            {
               log.warn("Could not set timeout (" + timeout + ") on http client transport as method not available with JDK 1.4 (only JDK 1.5 or higher)");
            }
            catch(IllegalAccessException e)
            {
               log.error("Error setting http client connection timeout.");
               log.debug(e);
            }
            catch(InvocationTargetException e)
            {
               log.error("Error setting http client connection timeout.");
               log.debug(e);
            }
         }
         catch(NumberFormatException e)
         {
            log.error("Could not set timeout for http client connection because value (" + timeoutValue + ") is not a number.");
         }
      }
   }

   private String validateURL(String url)
   {
      String validatedUrl = url;

      if(validatedUrl.startsWith("servlet"))
      {
         // servlet:// is a valid protocol, but only in the remoting world, so need to convert to http
         validatedUrl = "http" + validatedUrl.substring("servlet".length());
      }
      return validatedUrl;
   }

   protected HttpURLConnection createURLConnection(String url, Map metadata) throws IOException
   {
      URL externalURL = null;
      HttpURLConnection httpURLConn = null;

      // need to find out if need to use a proxy or not
      String proxyHost = null;
      String proxyportString = null;
      int proxyPort = -1;
      boolean proxyOn = true;

      if(metadata != null)
      {
         // first check the metadata as will have precedence
         proxyHost = (String) metadata.get("http.proxyHost");
         proxyportString = (String) metadata.get("http.proxyPort");
         if(proxyportString != null && proxyportString.length() > 0)
         {
            try
            {
               proxyPort = Integer.parseInt(proxyportString);
            }
            catch(NumberFormatException e)
            {
               log.warn("Error converting proxy port specified (" + proxyportString + ") to a number.");
            }
         }
      }

      // if not set in the metadata, try the system properties
      if(proxyHost == null)
      {
         // have found that the actual system property to be set can differ between VMs,
         // so will try all the ones I know of.
         proxyHost = System.getProperty("http.proxyHost");
         if(proxyHost == null)
         {
            proxyHost = System.getProperty("proxyHost");
         }

         // since not set by metadata, need to check if proxy turned off by property setting
         String proxyOnString = System.getProperty("http.proxySet");
         if(proxyOnString == null)
         {
            proxyOnString = System.getProperty("proxySet");
         }
         if(proxyOnString != null)
         {
            proxyOn = Boolean.getBoolean(proxyOnString);
         }
      }
      if(proxyPort < 0)
      {
         String proxyPortString = System.getProperty("http.proxyPort");
         if(proxyPortString == null)
         {
            proxyPortString = System.getProperty("proxyPort");
         }
         if(proxyPortString != null)
         {
            try
            {
               proxyPort = Integer.parseInt(proxyPortString);
            }
            catch(NumberFormatException e)
            {
               log.warn("Error converting proxy port specified (" + proxyportString + ") to a number.");
            }
         }
      }

      // now determin if going to use proxy or not
      if(proxyHost != null && proxyOn)
      {
         // is ok if port is -1 since URL constructor will use default port for protocol
         externalURL = new URL("http", proxyHost, proxyPort, url);

         httpURLConn = (HttpURLConnection) externalURL.openConnection();

         // since know it is a proxy being used, see if have proxy auth
         String proxyAuth = getProxyAuth(metadata);
         if(proxyAuth != null)
         {
            httpURLConn.setRequestProperty("Proxy-Authorization", proxyAuth);
         }
      }
      else
      {
         externalURL = new URL(url);
         httpURLConn = (HttpURLConnection) externalURL.openConnection();
      }

      return httpURLConn;
   }

   private String getProxyAuth(Map metadata)
   {
      String authString = null;
      String username = null;
      String password = null;

      if(metadata != null)
      {
         username = (String) metadata.get("http.proxy.username");
      }
      if(username == null || username.length() == 0)
      {
         username = System.getProperty("http.proxy.username");
      }
      if(metadata != null)
      {
         password = (String) metadata.get("http.proxy.password");
      }
      if(password == null || password.length() == 0)
      {
         password = System.getProperty("http.proxy.password");
      }

      if(username != null && password != null)
      {
         StringBuffer buffer = new StringBuffer();
         buffer.append(username);
         buffer.append(":");
         buffer.append(password);

         String encoded = Base64.encodeBytes(buffer.toString().getBytes());

         authString = "Basic " + encoded;

      }

      return authString;
   }

   private String getBasicAuth(Map metadata)
   {
      String authString = null;
      String username = null;
      String password = null;

      if(metadata != null)
      {
         username = (String) metadata.get("http.basic.username");
      }
      if(username == null || username.length() == 0)
      {
         username = System.getProperty("http.basic.username");
      }
      if(metadata != null)
      {
         password = (String) metadata.get("http.basic.password");
      }
      if(password == null || password.length() == 0)
      {
         password = System.getProperty("http.basic.password");
      }

      if(username != null && password != null)
      {
         StringBuffer buffer = new StringBuffer();
         buffer.append(username);
         buffer.append(":");
         buffer.append(password);

         String encoded = Base64.encodeBytes(buffer.toString().getBytes());

         authString = "Basic " + encoded;

      }

      return authString;
   }


   /**
    * subclasses must implement this method to provide a hook to connect to the remote server, if this applies
    * to the specific transport. However, in some transport implementations, this may not make must difference since
    * the connection is not persistent among invocations, such as SOAP.  In these cases, the method should
    * silently return without any processing.
    *
    * @throws org.jboss.remoting.ConnectionFailedException
    *
    */
   protected void handleConnect() throws ConnectionFailedException
   {
      // NO OP as not statefull connection
   }

   /**
    * subclasses must implement this method to provide a hook to disconnect from the remote server, if this applies
    * to the specific transport. However, in some transport implementations, this may not make must difference since
    * the connection is not persistent among invocations, such as SOAP.  In these cases, the method should
    * silently return without any processing.
    */
   protected void handleDisconnect()
   {
      // NO OP as not statefull connection
   }

   /**
    * Each implementation of the remote client invoker should have
    * a default data type that is uses in the case it is not specified
    * in the invoker locator uri.
    *
    * @return
    */
   protected String getDefaultDataType()
   {
      return HTTPMarshaller.DATATYPE;
   }

}