/*
* JBoss, Home of Professional Open Source
* Copyright 2005, JBoss Inc., and individual contributors as indicated
* by the @authors tag. See the copyright.txt in the distribution for a
* full listing of individual contributors.
*
* This is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2.1 of
* the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free
* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
* 02110-1301 USA, or see the FSF site: http://www.fsf.org.
*/

package org.jboss.remoting.transport.multiplex;

import java.io.BufferedInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InterruptedIOException;
import java.io.OutputStream;
import java.net.Socket;
import java.util.Iterator;

import org.jboss.logging.Logger;
import org.jboss.remoting.transport.multiplex.utility.StoppableThread;

/**
 * 
 * <p>
 * Copyright (c) 2005
 * <p>
 * @author <a href="mailto:r.sigal@computer.org">Ron Sigal</a>
 */
public class InputMultiplexor
{
   protected static final Logger log = Logger.getLogger(InputMultiplexor.class);
   

/**
 *
 * @return
 */
   static InputThread getAnInputThread(MultiplexingManager manager, Socket socket, OutputStream os) throws IOException
   {
      return new InputThread(manager, socket, os);
   }



   public static class InputThread extends StoppableThread
   {
      private static final Logger log = Logger.getLogger(InputMultiplexor.InputThread.class);
      
      private static final int READING_DATA = 0;
      private static final int SAW_MARKER = 1;
      private static final int READING_ID = 2;
      
      private InputStream is;
      private OutputStream currentOutputStream;
      private ByteArrayOutputStream bos;
      private MultiplexingManager manager;
      private int state;
      private boolean managerShuttingDown = false;
      private int dataInCount = 0;
      private SocketId currentSocketId;
      private boolean tracing;
      
      
      public InputThread(MultiplexingManager manager, Socket socket, OutputStream os) throws IOException
      {
         this.is = new BufferedInputStream(socket.getInputStream());
         this.manager = manager;
         
         currentOutputStream = os;
         bos = new ByteArrayOutputStream();
         int state = READING_DATA;
         tracing = log.isTraceEnabled();
      }
      
      
      public synchronized void shutdown()
      {
         // in case thread is still reading
         super.shutdown();
         interrupt();
         
         // in case thread is in stayAlive()
         managerShuttingDown = true;
         notify();
      }
      
      
      /**
       * 
       */
      protected void doInit()
      {
         log.debug("input thread starting");
      }
      
      
      /**
       * 
       */
      protected void doRun()
      {
         int b = 0;
         
         try
         {
            b = is.read();
            
            while (b >= 0 && isRunning())
            {
               if (tracing)
                  log.trace("" +( 0xff &  b));
                  
               switch(state)
               {
               
               case READING_DATA:
                  
                  if (b == OutputMultiplexor.MARKER)
                     state = SAW_MARKER;
                  else
                  {
                     try
                     {
                        currentOutputStream.write(b);
                     }
                     catch (IOException e)
                     {
                        log.debug("unable to write to PipedOutputStream (" 
                              + currentSocketId.getPort()
                              + "): reading thread must have died", e);
                        currentOutputStream = manager.getOutputStreamByLocalSocket(SocketId.DEADLETTER_SOCKET_ID);
                     }
                     
                     dataInCount++;
                  }
                  
                  break;
                  
                  
               case SAW_MARKER:
                  
                  if (b == OutputMultiplexor.MARKER)
                  {
                     state = READING_DATA;
                     
                     try
                     {
                        currentOutputStream.write(b);
                     }
                     catch (IOException e)
                     {
                        // OK: probably InputStream shut down on virtual socket
                        log.debug("unable to write to PipedOutputStream (" + currentSocketId.getPort() + ")", e);
                        currentOutputStream = manager.getOutputStreamByLocalSocket(SocketId.DEADLETTER_SOCKET_ID);
                     }

                     dataInCount++;
                  }
                  else
                  {
                     state = READING_ID;
                     currentOutputStream.flush();
                     bos.reset();
                     bos.write(b);
                  }
                  
                  break;
                  
               case READING_ID:
                  
                  if (b == OutputMultiplexor.MARKER)
                  {
                     state = READING_DATA;
                     currentSocketId = new SocketId(bos.toByteArray());
                     bos.reset();
                     currentOutputStream = manager.getOutputStreamByLocalSocket(currentSocketId);
                     
                     if (tracing)
                        log.trace("redirecting input to socket: " + currentSocketId.getPort());
                     
                     if (currentOutputStream == null)
                     {
                        log.error("unknown socket id: " + currentSocketId.getPort());
                        currentOutputStream = manager.getOutputStreamByLocalSocket(SocketId.DEADLETTER_SOCKET_ID);
                     } 
                  }
                  else
                     bos.write(b);
                  
                  break; 
                  
               } // switch
               
               b = is.read();
               
            } // while
            
         } // try
         
         catch (InterruptedIOException e)
         {
            handleError("interrupted", e);
         }
         catch (IOException e)
         {
            handleError("i/o error", e);
         }
         
         super.shutdown();
         
         log.debug("last byte: " + b);
         log.debug("isRunning: " + isRunning());
      }
      
      
      protected void doShutDown()
      {  
         //            TODO: work this out
         //            if (terminatedOnError)
         //               multiplexor.shutdown();
         
         log.debug("input thread: data bytes read: " + dataInCount);
         
         flushPipes();
         stayAlive();
         
         log.debug("input thread shutting down");
      }
      
      
      /**
       *
       * @param error
       * @param e
       */
      protected void handleError(String error, Throwable e)
      {
         if (e instanceof InterruptedException || e instanceof IOException)
         {
            if (tracing)
               log.debug(error, e);
         }
         else
            log.debug(error, e);
         
         //    TODO
         //            running = false;
         //            terminatedOnError = true;
      }
      
      
      protected synchronized void stayAlive()
      {
         log.debug("input thread entering stayAlive()");
         
         while (!managerShuttingDown)
         {
            try
            {
               wait();
            }
            catch (InterruptedException e)
            {
               log.debug("stayAlive(): InterruptedException");
            }
            
            log.debug("stayAlive(): wait ended");
         }
      }
      
      
      protected void flushPipes()
      {
         Iterator it = manager.getAllOutputStreams().iterator();
         
         while (it.hasNext())
         {
            OutputStream os = (OutputStream) it.next();
            
            try
            {
               os.flush();
//             os.close();
            }
            catch (IOException ignored)
            {
               log.error("unable to flush a PipedOutputStream");
            }
         }

      }
   }
}
