/*
* JBoss, Home of Professional Open Source
* Copyright 2005, JBoss Inc., and individual contributors as indicated
* by the @authors tag. See the copyright.txt in the distribution for a
* full listing of individual contributors.
*
* This is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2.1 of
* the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free
* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
* 02110-1301 USA, or see the FSF site: http://www.fsf.org.
*/

/*
 * Created on Jul 24, 2005
 */
 
package org.jboss.remoting.transport.multiplex;

import java.io.DataOutputStream;
import java.io.IOException;
import java.io.Serializable;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketAddress;
import java.net.SocketException;
import java.net.SocketTimeoutException;
import java.nio.channels.ServerSocketChannel;

import javax.net.ServerSocketFactory;
import javax.net.ssl.SSLServerSocket;

import org.jboss.logging.Logger;


/**
 * 
 * <p>
 * Copyright (c) 2005
 * <p>
 * @author <a href="mailto:r.sigal@computer.org">Ron Sigal</a>
 */
public class MasterServerSocket extends ServerSocket implements Serializable
{
   private static final Logger log = Logger.getLogger(MasterServerSocket.class);
   private ServerSocket ss;

   
//////////////////////////////////////////////////////////////////////////////////////////////////
///        The following constructors duplicate those of the parent class ServerSocket        '///
//////////////////////////////////////////////////////////////////////////////////////////////////
   
/**
 * @throws java.io.IOException
 */
   public MasterServerSocket() throws IOException
   {
      ss = new ServerSocket();
   }

/**
 * @param port
 * @throws java.io.IOException
 */
   public MasterServerSocket(int port) throws IOException
   {
      ss = new ServerSocket(port);
   }

/**
 * @param port
 * @param backlog
 * @throws java.io.IOException
 */
   public MasterServerSocket(int port, int backlog) throws IOException
   {
      ss = new ServerSocket(port, backlog);
   }

/**
 * @param port
 * @param backlog
 * @param bindAddr
 * @throws java.io.IOException
 */
   public MasterServerSocket(int port, int backlog, InetAddress bindAddr) throws IOException
   {
      ss = new ServerSocket(port, backlog, bindAddr);
   }
   
   
/**
 * @param ssf
 * @throws IOException
 */
   public MasterServerSocket(ServerSocketFactory ssf) throws IOException
   {
      ss = ssf.createServerSocket();
   }


/**
 * @param ssf
 * @param port
 * @throws IOException
 */
   public MasterServerSocket(ServerSocketFactory ssf, int port) throws IOException
   {
      ss = ssf.createServerSocket(port);
   }


/**
 * @param ssf
 * @param port
 * @param backlog
 * @throws IOException
 */
   public MasterServerSocket(ServerSocketFactory ssf, int port, int backlog) throws IOException
   {
      ss = ssf.createServerSocket(port, backlog);
   }


/**
 * @param ssf
 * @param port
 * @param backlog
 * @param bindAddr
 * @throws IOException
 */
   public MasterServerSocket(ServerSocketFactory ssf, int port, int backlog, InetAddress bindAddr)
   throws IOException
   {
      ss = ssf.createServerSocket(port, backlog, bindAddr);
   }


   //////////////////////////////////////////////////////////////////////////////////////////////////
   ///                  The following methods are required of any ServerSocket                   '///
   //////////////////////////////////////////////////////////////////////////////////////////////////


   /*
     ok: public Socket              accept() throws IOException;
     ok: public void                bind(SocketAddress endpoint) throws IOException;
     ok: public void                bind(SocketAddress endpoint, int backlog) throws IOException;
     ok: public void                close(F) throws IOException;
     ok: public ServerSocketChannel getChannel();
     ok: public InetAddress         getInetAddress();
     ok: public int                 getLocalPort();
     ok: public SocketAddress       getLocalSocketAddress();
     ok: public int                 getReceiveBufferSize() throws SocketException;
     ok: public boolean             getReuseAddress() throws SocketException;
     ok: public int                 getSoTimeout() throws IOException;
     ok: public boolean             isBound();
     ok: public boolean             isClosed();
     ok: public void                setReceiveBufferSize(int size) throws SocketException;
     ok: public void                setReuseAddress(boolean on) throws SocketException;
     ok: public void                setSoTimeout(int timeout) throws SocketException;
     ok: public String              toString();
   */

/**
 *
 */
   public Socket accept() throws IOException
   {
      long start = System.currentTimeMillis();
      int timeout = getSoTimeout();
      int savedTimeout = timeout;
      int timeLeft = 0;
      Socket socket = null;

      while (true)
      {
         if (timeout > 0)
            if ((timeLeft = timeout - (int) (System.currentTimeMillis() - start)) <= 0)
               throw new SocketTimeoutException("Accept timed out");

         setSoTimeout(timeLeft);

         try
         {
            socket = ss.accept();
         }
         finally
         {
            setSoTimeout(savedTimeout);
         }

         MultiplexingManager manager = MultiplexingManager.getaManager(socket);
         MultiplexingDataInputStream is = null;
         Protocol protocol = null;
         SocketId clientPort = null;

         try
         {
            is = new MultiplexingDataInputStream(manager.getAnInputStream(SocketId.SERVER_SOCKET_ID, null));
            protocol = manager.getProtocol();

            if (timeout > 0)
               if ((timeLeft = timeout - (int) (System.currentTimeMillis() - start)) <= 0)
                  throw new SocketTimeoutException("Accept timed out");

            clientPort = protocol.acceptConnect(is, timeLeft);
         }
         catch (IOException e)
         {
            log.error("i/o exception in MasterServerSocket.accept()");
            manager.decrementReferences();
            throw e;
         }

         log.debug("accept(): clientPort:  " + clientPort.getPort());

         // connection from independent VirtualServerSocket
         if (clientPort.getPort() < 0)
         {
            MultiplexingOutputStream mos = new MultiplexingOutputStream(manager, SocketId.SERVER_SOCKET_CONNECT_ID);
            DataOutputStream os = new DataOutputStream(mos);
            MultiplexingDataInputStream mdis = null;

            try
            {
               protocol.answerConnect(os, SocketId.SERVER_SOCKET_CONNECT_PORT);

               // Make sure that the remote VirtualServerSocket is registered with the
               // local MultiplexingManager before continuing. This check avoids a race condition
               // in which local VirtualSockets try to connect with the remote VirtualServerSocket
               // before the local MultiplexingManager knows it exists.
               MultiplexingInputStream mis = manager.getAnInputStream(SocketId.SERVER_SOCKET_VERIFY_ID, null);
               mdis = new MultiplexingDataInputStream(mis);

               if (timeout > 0)
                  if ((timeLeft = timeout - (int) (System.currentTimeMillis() - start)) <= 0)
                     throw new SocketTimeoutException("Accept timed out");

               protocol.getConnectVerification(mdis, timeLeft);
            }
            catch (IOException e)
            {
               // If this connect timed out at the other end, there may not be an OutputStream to write to.
               log.error("unable to respond to connect request");
               manager.decrementReferences();

               if (e instanceof SocketTimeoutException)
                  throw new SocketTimeoutException("Accept timed out");

               throw e;
            }

            // Keept trying to accept a request for a VirtualSocket connection.
            continue;
         }

         VirtualSocket virtualSocket = null;

         try
         {
            virtualSocket = new VirtualSocket(manager, clientPort);
         }
         catch (IOException e)
         {
            manager.decrementReferences();
            throw e;
         }

         // Make sure that the local VirtualSocket is registered before confirming the connection
         // to the remote VirtualSocket.  This check avoids a race condition in which the remote
         // VirtualSocket starts sending bytes before the local VirtualSocket's MultiplexingInputStream
         // is registered.
//         while (!manager.isSocketRegistered(virtualSocket.getLocalSocketId()))
//            try
//            {
//               Thread.sleep(100);
//            }
//         	catch (InterruptedException ignored)
//         	{
//         	   log.info("interruption: waiting for manager.isSocketRegistered(clientPort)");
//         	}

         int localPort = virtualSocket.getLocalVirtualPort();

         try
         {
            protocol.answerConnect(new DataOutputStream(virtualSocket.getOutputStream()), localPort);
         }
         catch (IOException e)
         {
            // If this connect timed out at the other end, there may not be an OutputStream to write to.
            log.error("unable to respond to connect request");
            virtualSocket.close();
            throw e;
         }

         MultiplexingDataInputStream mdis = null;

         try
         {
            // Make sure that the remote VirtualSocket is registered before returning the local VirtualSocket.
            // This check avoids a race condition in which the local VirtualSocket starts sending bytes
            // before the remote VirtualSocket's MultiplexingInputStream is registered.
            mdis = new MultiplexingDataInputStream((MultiplexingInputStream) virtualSocket.getInputStream());

            if (timeout > 0)
               if ((timeLeft = timeout - (int) (System.currentTimeMillis() - start)) <= 0)
                  throw new SocketTimeoutException("Accept timed out");

            protocol.getConnectVerification(mdis, timeLeft);
         }
         catch (IOException e)
         {
            log.error("unable to get confirmation that connect was successful");
            virtualSocket.close();

            if (e instanceof SocketTimeoutException)
               throw new SocketTimeoutException("Accept timed out");

            throw e;
         }

         return virtualSocket;
      }
   }


/**
 *
 */
   public void bind(SocketAddress endpoint) throws IOException
   {
      ss.bind(endpoint);
   }


/**
 *
 */
   public void bind(SocketAddress endpoint, int backlog) throws IOException
   {
      ss.bind(endpoint, backlog);
   }


/**
 *
 */
   public void close() throws IOException
   {
      log.debug("MasterServerSocket: closing");
      ss.close();
   }

/**
 *
 */
   public ServerSocketChannel getChannel()
   {
      return ss.getChannel();
   }


/**
 *
 */
   public InetAddress getInetAddress()
   {
      return ss.getInetAddress();
   }


/**
 *
 */
   public int getLocalPort()
   {
      return ss.getLocalPort();
   }


/**
 *
 */
   public SocketAddress getLocalSocketAddress()
   {
      return ss.getLocalSocketAddress();
   }


/**
 *
 */
   public int getReceiveBufferSize() throws SocketException
   {
      return ss.getReceiveBufferSize();
   }


/**
 *
 */
   public boolean getReuseAddress() throws SocketException
   {
      return ss.getReuseAddress();
   }


/**
 *
 */
   public int getSoTimeout() throws IOException
   {
      return ss.getSoTimeout();
   }


/**
 *
 */
   public boolean isBound()
   {
      return ss.isBound();
   }


/**
 *
 */
   public boolean isClosed()
   {
      return ss.isClosed();
   }


/**
 *
 */
   public void setReceiveBufferSize(int size) throws SocketException
   {
      ss.setReceiveBufferSize(size);
   }


/**
 *
 */
   public void setReuseAddress(boolean on) throws SocketException
   {
      ss.setReuseAddress(on);
   }


/**
 *
 */
   public void setSoTimeout(int timeout) throws SocketException
   {
      ss.setSoTimeout(timeout);
   }


/**
 *
 */
   public String toString()
   {
  	if (!isBound())
  	    return "MasterServerSocket[unbound]";

  	return "MasterServerSocket[addr=" + getInetAddress() +  ",localport=" + getLocalPort()  + "]";
   }


//////////////////////////////////////////////////////////////////////////////////////////////////
///              The following methods are specific to MasterServerSocket           '///
//////////////////////////////////////////////////////////////////////////////////////////////////

   public int acceptServerSocketConnection() throws IOException
   {
      long start = System.currentTimeMillis();
      int timeout = getSoTimeout();
      int savedTimeout = timeout;
      int timeLeft = 0;

      Socket socket = ss.accept();
      MultiplexingManager manager = MultiplexingManager.getaManager(socket);
      MultiplexingDataInputStream is = null;
      Protocol protocol = null;
      SocketId clientPort = null;

      try
      {
         is = new MultiplexingDataInputStream(manager.getAnInputStream(SocketId.SERVER_SOCKET_ID, null));
         protocol = manager.getProtocol();

         if (timeout > 0)
            if ((timeLeft = timeout - (int) (System.currentTimeMillis() - start)) <= 0)
               throw new SocketTimeoutException("Accept timed out");

         clientPort = protocol.acceptConnect(is, timeLeft);
      }
      catch (IOException e)
      {
         log.error("i/o exception in MasterServerSocket.acceptServerSocketConnection()", e);
         manager.decrementReferences();

         if (e instanceof SocketTimeoutException)
            throw new SocketTimeoutException("Accept timed out");

         throw e;
      }

      if (clientPort.getPort() != SocketId.SERVER_SOCKET_PORT)
      {
         manager.decrementReferences();
         String message = "received connect request not from a VirtualServerSocket";
         log.error(message);
         throw new IOException(message);
      }

      DataOutputStream os = new DataOutputStream(new MultiplexingOutputStream(manager, SocketId.SERVER_SOCKET_CONNECT_ID));

      try
      {
         protocol.answerConnect(os, SocketId.SERVER_SOCKET_CONNECT_PORT);

         // Make sure that the remote VirtualServerSocket is registered with the
         // local MultiplexingManager before continuing. This check avoids a race condition
         // in which local VirtualSocket try to connect with the remote VirtualServerSocket
         // before the local MultiplexingManager knows it exists.
         MultiplexingInputStream mis = manager.getAnInputStream(SocketId.SERVER_SOCKET_VERIFY_ID, null);
         MultiplexingDataInputStream mdis = new MultiplexingDataInputStream(mis);

         if (timeout > 0)
            if ((timeLeft = timeout - (int) (System.currentTimeMillis() - start)) <= 0)
               throw new SocketTimeoutException("Accept timed out");

         protocol.getConnectVerification(mdis, timeLeft);
      }
      catch (IOException e)
      {
         // If this connect timed out at the other end, there may not be an OutputStream to write to.
         log.error("unable to respond to connect request");
         manager.decrementReferences();
         throw e;
      }

      return manager.getSocket().getLocalPort();
   }
}