/*
* JBoss, Home of Professional Open Source
* Copyright 2005, JBoss Inc., and individual contributors as indicated
* by the @authors tag. See the copyright.txt in the distribution for a
* full listing of individual contributors.
*
* This is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2.1 of
* the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free
* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
* 02110-1301 USA, or see the FSF site: http://www.fsf.org.
*/

package org.jboss.remoting.transport.multiplex;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.InetSocketAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketException;
import java.net.SocketTimeoutException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.net.SocketFactory;

import org.jboss.logging.Logger;
import org.jboss.remoting.transport.multiplex.utility.GrowablePipedOutputStream;
import org.jboss.remoting.transport.multiplex.utility.StoppableThread;


/**
 * <p>
 * Copyright (c) 2005
 * <p>
 * @author <a href="mailto:r.sigal@computer.org">Ron Sigal</a>
 */
public class MultiplexingManager
{
   private static final Logger log = Logger.getLogger(MultiplexingManager.class);

   /**
    * This object is used to synchronized operations on the managersByRemoteAddress map.
    */
   private static Object shareableMapLock = new Object();

   /** A HashMap<InetSocketAddress, HasnSet<MultiplexingManager>> of sets of MultiplexingManager's
    *  indexed by remote address.  Holds all MultiplexingManagers whose peer has an attached
    *  VirtualServerSocket. */
   private static Map shareableManagers = new HashMap();

   /** This object is used to synchronized operations on the managersByRemoteAddress map. */
   private static Object localAddressMapLock = new Object();

   /** A HashMap<InetSocketAddress, HasnSet<MultiplexingManager>> of sets of MultiplexingManager's
    *  indexed by local address */
   private static Map managersByLocalAddress = new HashMap();

   /** This object is used to synchronized operations on the managersByRemoteAddress map. */
   private static Object remoteAddressMapLock = new Object();

   /** A HashMap<InetSocketAddress, HasnSet<MultiplexingManager>> of sets of MultiplexingManager's
    *  indexed by remote address */
//   private static Map managersByRemoteAddress = Collections.synchronizedMap(new HashMap());
   private static Map managersByRemoteAddress = new HashMap();

   /** OutputMultiplexor associated with this MultiplexingManager */
   private OutputMultiplexor outputMultiplexor;

   /** A HashMap<SocketId, VirtualSocket> of VirtualSocket's indexed by local SocketId */
   private Map socketMap = Collections.synchronizedMap(new HashMap());

   /** A HashSet of SocketId's registered on this MultiplexingManager */
   private Set registeredSockets = Collections.synchronizedSet(new HashSet());

   /** A HashMap<Long, OutputStream> of OutputStream's associated with virtual sockets  */
   private Map outputStreamMap = Collections.synchronizedMap(new HashMap());

   /** A HashMap<Long, InputStream> of InputStream's associated with virtual sockets  */
   private Map inputStreamMap = Collections.synchronizedMap(new HashMap());

   /** Threads waiting to be notified of registration of remote server socket */
   private Set threadsWaitingForRemoteServerSocket = new HashSet();

   /** Protocol object for handling connection/disconnection communications */
   private Protocol protocol;

   /** Actual local socket upon which this family of virtual sockets is based */
   private Socket socket;

   /** bound state of actual local socket */
   private boolean bound = false;

   /** connected state of actual local socket */
   private boolean connected = false;

   /** SocketAddress of remote actual socket to which this Manager's socket is connected */
   private InetSocketAddress remoteSocketAddress;

   /** SocketAddress to which this manager's actual socket is bound */
   private InetSocketAddress localSocketAddress;

   /** Represents local port on which this manager's actual socket is bound, with any local address */
   private InetSocketAddress localWildCardAddress;

   /** Currently registered server socket */
   private ServerSocket serverSocket;

   /** Indicates if remote server socket has been registered */
   private boolean remoteServerSocketRegistered = false;

   /** InputThread reading from socket's InputStream and distributing to virtual sockets */
   private InputMultiplexor.InputThread inputThread;

   /** Thread for writing to socket's OutputStream */
   private OutputMultiplexor.OutputThread outputThread;

   /** Thread for getting asynchronous messages from remote Protocol */
   private Protocol.BackChannelThread backChannelThread;

   /** OutputStream for unknown virtual sockets */
   private OutputStream deadLetterOutputStream = new ByteArrayOutputStream();

   /** Manages the shutdown handshaking protocol between peer MultiplexingManagers */
   private ShutdownManager shutdownManager = new ShutdownManager();

   /** true if and only MultiplexingManager is completely shut down */
   private boolean shutdown = false;

   /** Holds PendingActions waiting to be performed */
   private List pendingActions = new ArrayList();

   /** Removes virtual sockets from closingSockets and closes them. */
   private PendingActionThread pendingActionThread;

   /** The logging trace level */
   private boolean tracing;

   private long id;

   /** SocketFactory to use for creating sockets */
   private SocketFactory socketFactory;



   protected MultiplexingManager(SocketFactory socketFactory) throws IOException
   {
      id = new Date().getTime();

      if (socketFactory != null)
      {
         this.socketFactory = socketFactory;
         socket = socketFactory.createSocket();
      }
      else
         socket = new Socket();

      try
      {
         socket.setReuseAddress(true);
      }
      catch (SocketException ignored)
      {
         log.error(ignored);
      }

      log.debug("new MultiplexingManager: " + socket.toString());
   }


/**
 *
 * @param socket
 * @throws IOException
 */
   protected MultiplexingManager(Socket socket) throws IOException
   {
      this.socket = socket;
      id = new Date().getTime();
      setup();
      log.debug("new MultiplexingManager: " + socket.toString());
   }


/**
 *
 * @param address
 * @param timeout TODO
 * @param socketFactory TODO
 * @throws IOException
 */
   protected MultiplexingManager(InetSocketAddress address, int timeout, SocketFactory socketFactory)
   throws IOException
   {
      if (socketFactory != null)
      {
         this.socketFactory = socketFactory;
         socket = socketFactory.createSocket();
      }
      else
         socket = new Socket();

      try
      {
         socket.setReuseAddress(true);
      }
      catch (SocketException ignored)
      {
         log.error(ignored);
      }

      socket.connect(address, timeout);
      setup();
      log.debug("new MultiplexingManager: " + socket.toString());
   }


/**
 *
 */
   protected synchronized void setup() throws IOException
   {
      // register dead letter output stream (for unrecognized destinations)
      outputStreamMap.put(SocketId.DEADLETTER_SOCKET_ID, deadLetterOutputStream);

      // TODO: what was this for???
      registeredSockets.add(SocketId.PROTOCOL_SOCKET_ID);
      registeredSockets.add(SocketId.SERVER_SOCKET_ID);
      registeredSockets.add(SocketId.SERVER_SOCKET_CONNECT_ID);
      registeredSockets.add(SocketId.SERVER_SOCKET_VERIFY_ID);
      registeredSockets.add(SocketId.BACKCHANNEL_SOCKET_ID);

      // set up standard piped streams
      getAnInputStream(SocketId.PROTOCOL_SOCKET_ID, null);
      getAnInputStream(SocketId.SERVER_SOCKET_ID, null);
      getAnInputStream(SocketId.SERVER_SOCKET_CONNECT_ID, null);
      getAnInputStream(SocketId.SERVER_SOCKET_VERIFY_ID, null);
      getAnInputStream(SocketId.BACKCHANNEL_SOCKET_ID, null);

      // start output thread
      outputMultiplexor = new OutputMultiplexor(this, socket);
      outputThread = outputMultiplexor.getAnOutputThread();
      outputThread.setName(outputThread.getName() + ":output(" + socket.toString() + ")");
//      outputThread.setDaemon(true);
      outputThread.start();

      // start input thread
      inputThread = InputMultiplexor.getAnInputThread(this, socket, deadLetterOutputStream);
      inputThread.setName(inputThread.getName() + ":input(" + socket.toString() + ")");
//      inputThread.setDaemon(true);
      inputThread.start();

      // start back channel thread
      protocol = new Protocol(this);
      backChannelThread = protocol.getBackChannelThread();
      backChannelThread.setName(backChannelThread.getName() + ":backchannel(" + socket.toString() + ")");
//      backChannelThread.setDaemon(true);
      backChannelThread.start();

      // start socket close thread
      pendingActionThread = new PendingActionThread();
      pendingActionThread.setName(pendingActionThread.getName() + ":background(" + socket.toString() + ")");
//      socketCloseThread.setPriority(socketCloseThread.getPriority() + 3);
//      pendingActionThread.setDaemon(true);
      pendingActionThread.start();

//      if (!bound)
//      {
//         bound = true;
//         registerByLocalAddress(new InetSocketAddress(socket.getLocalAddress(), socket.getLocalPort()));
//      }

      registerByLocalAddress(new InetSocketAddress(socket.getLocalAddress(), socket.getLocalPort()));
      registerByRemoteAddress(new InetSocketAddress(socket.getInetAddress(), socket.getPort()));
      bound = true;
      connected = true;

      tracing = log.isTraceEnabled();
   }


/**
 *
 * @param socket
 * @return
 * @throws IOException
 *
 * TODO: what if multiplexor already exists?
 */
   public static MultiplexingManager getaManager(Socket socket) throws IOException
   {
      log.debug("entering getaManager(Socket socket)");
      return new MultiplexingManager(socket);
   }


/**
 * @param address
 * @return
 * @throws IOException
 */
   public static synchronized MultiplexingManager
   getaManagerByLocalAddress(InetSocketAddress address) throws IOException
   {
      return getaManagerByLocalAddress(address, null);
   }


/**
 *
 * @param address
 * @param socketFactory
 * @return
 * @throws IOException
 */
   public static synchronized MultiplexingManager
   getaManagerByLocalAddress(InetSocketAddress address, SocketFactory socketFactory) throws IOException
   {
      log.debug("entering getaManagerByLocalAddress(InetSocketAddress address)");
      MultiplexingManager m = null;

      synchronized (localAddressMapLock)
      {
         HashSet managers = (HashSet) managersByLocalAddress.get(address);

         if (managers != null)
         {
            Iterator it = managers.iterator();

            while (it.hasNext())
            {
               m = (MultiplexingManager) it.next();

               if (socketFactory == m.getSocketFactory())
               {
                  try
                  {
                     m.shutdownManager.incrementReferences();
                     return m;
                  }
                  catch (IOException e)
                  {
                  }
               }
            }
         }
      }

      log.debug("There is no joinable MultiplexingManager. Creating new one.");
      m = new MultiplexingManager(socketFactory);
      m.bind(address);
      return m;
   }


/**
 *
 * @param address
 * @param timeout
 * @return
 * @throws IOException
 */
   public static synchronized MultiplexingManager
   getaManagerByRemoteAddress(InetSocketAddress address, int timeout) throws IOException
   {
      return getaManagerByRemoteAddress(address, timeout, null);
   }


/**
 *
 * @param address
 * @param timeout
 * @param socketFactory
 * @return
 * @throws IOException
 */
   public static synchronized MultiplexingManager
   getaManagerByRemoteAddress(InetSocketAddress address, int timeout, SocketFactory socketFactory)
   throws IOException
   {
      log.debug("entering getaManagerByRemoteAddress(InetSocketAddress address)");

//      if (isOnServer())
//      {

      // Check each of the MultiplexingManagers connected to the target remote address, looking
      // for one which is not shutting down.
      synchronized(remoteAddressMapLock)
      {
         HashSet managers = (HashSet) managersByRemoteAddress.get(address);

         if (managers != null && !managers.isEmpty())
         {
            Iterator it = managers.iterator();

            while (it.hasNext())
            {
               MultiplexingManager m = (MultiplexingManager) it.next();

               if (socketFactory == m.getSocketFactory())
               {
                  try
                  {
                     m.shutdownManager.incrementReferences();
                     return m;
                  }
                  catch (Exception e)
                  {
                     log.debug("manager shutting down: " + m);
                  }
               }
            }
         }
      }

      return new MultiplexingManager(address, timeout, socketFactory);
   }



/**
 * @param remoteAddress
 * @param localAddress
 * @param timeout
 * @return
 * @throws IOException
 */
   public static synchronized MultiplexingManager
   getaManagerByAddressPair(InetSocketAddress remoteAddress, InetSocketAddress localAddress, int timeout)
   throws IOException
   {
      return getaManagerByAddressPair(remoteAddress, localAddress, timeout, null);
   }


/**
 * @param remoteAddress
 * @param localAddress
 * @param timeout
 * @param socketFactory
 * @return
 * @throws IOException
 */
   public static synchronized MultiplexingManager
   getaManagerByAddressPair(InetSocketAddress remoteAddress, InetSocketAddress localAddress,
                            int timeout, SocketFactory socketFactory)
   throws IOException
   {
      log.debug("entering getaManagerByRemoteAddress(InetSocketAddress address)");
      log.debug(localAddress);
      log.debug(remoteAddress);

      MultiplexingManager m;

      // Check each of the MultiplexingManagers connected to the target remote address, looking
      // for one which is not shutting down.
      synchronized(remoteAddressMapLock)
      {
         HashSet managers = (HashSet) managersByRemoteAddress.get(remoteAddress);

         if (managers != null && !managers.isEmpty())
         {
            Iterator it = managers.iterator();

            while (it.hasNext())
            {
               m = (MultiplexingManager) it.next();

               if (socketFactory == m.getSocketFactory())
               {
                  if (m.getSocket().getLocalAddress().equals(localAddress.getAddress()) &&
                        m.getSocket().getLocalPort() == localAddress.getPort())
                  {
                     try
                     {
                        m.shutdownManager.incrementReferences();
                        return m;
                     }
                     catch (Exception e)
                     {
                        log.debug("manager shutting down: " + m);
                     }
                  }
               }
            }
         }
      }

      log.debug("There is no joinable MultiplexingManager. Creating new one.");
      m = new MultiplexingManager(socketFactory);
      m.bind(localAddress);
//      m.connect(remoteAddress, timeout);
      return m;
   }


/**
 * @param address
 * @param timeout
 * @return
 * @throws IOException
 */
   public static synchronized MultiplexingManager
   getaShareableManager(InetSocketAddress address, int timeout) throws IOException
   {
      return getaShareableManager(address, timeout, null);
   }


/**
 * @param address
 * @param timeout
 * @param socketFactory
 * @return
 * @throws IOException
 */
   public static synchronized MultiplexingManager
   getaShareableManager(InetSocketAddress address, int timeout, SocketFactory socketFactory) throws IOException
   {
      log.debug("entering getaShareableManager(InetSocketAddress address)");

      // Check each of the shareable MultiplexingManagers connected to the target remote address, looking
      // for one which is not shutting down.
      synchronized(shareableMapLock)
      {
         HashSet managers = (HashSet) shareableManagers.get(address);

         if (managers != null && !managers.isEmpty())
         {
            Iterator it = managers.iterator();

            while (it.hasNext())
            {
               MultiplexingManager m = (MultiplexingManager) it.next();

               if (socketFactory == m.getSocketFactory())
               {
                  try
                  {
                     m.shutdownManager.incrementReferences();
                     return m;
                  }
                  catch (Exception e)
                  {
                     log.debug("manager shutting down: " + m);
                  }
               }
            }
         }
      }

      return new MultiplexingManager(address, timeout, socketFactory);
   }


/**
 * @param address
 * @param timeout
 * @param socketFactory
 * @return
 * @throws IOException
 */
   public static  MultiplexingManager
   getAnExistingShareableManager(InetSocketAddress address, SocketFactory socketFactory)
   throws IOException
   {
      log.debug("entering getAnExistingShareableManager()");

      // Check each of the shareable MultiplexingManagers connected to the target remote address, looking
      // for one which is not shutting down.
      synchronized(shareableMapLock)
      {
         HashSet managers = (HashSet) shareableManagers.get(address);

         if (managers != null && !managers.isEmpty())
         {
            Iterator it = managers.iterator();

            while (it.hasNext())
            {
               MultiplexingManager m = (MultiplexingManager) it.next();

               if (socketFactory == m.getSocketFactory())
               {
                  try
                  {
                     m.shutdownManager.incrementReferences();
                     return m;
                  }
                  catch (Exception e)
                  {
                     log.debug("manager shutting down: " + m);
                  }
               }
            }
         }
      }

      return null;
   }


/**
 * @param remoteAddress
 * @param localAddress
 * @param timeout
 * @return
 * @throws IOException
 */
   public static synchronized MultiplexingManager
   getaShareableManagerByAddressPair(InetSocketAddress remoteAddress, InetSocketAddress localAddress, int timeout)
   throws IOException
   {
      return getaShareableManagerByAddressPair(remoteAddress, localAddress, timeout, null);
   }


/**
 * @param remoteAddress
 * @param localAddress
 * @param timeout
 * @param socketFactory
 * @return
 * @throws IOException
 */
   public static synchronized MultiplexingManager
   getaShareableManagerByAddressPair(InetSocketAddress remoteAddress, InetSocketAddress localAddress,
                                     int timeout, SocketFactory socketFactory)
   throws IOException
   {
      log.debug("entering getaShareableManager(InetSocketAddress address)");
      MultiplexingManager m;

      // Check each of the shareable MultiplexingManagers connected to the target remote address, looking
      // for one which is not shutting down.
      synchronized(shareableMapLock)
      {
         HashSet managers = (HashSet) shareableManagers.get(remoteAddress);

         if (managers != null && !managers.isEmpty())
         {
            Iterator it = managers.iterator();

            while (it.hasNext())
            {
               m = (MultiplexingManager) it.next();

               if (socketFactory == m.getSocketFactory())
               {
                  if (m.getSocket().getLocalAddress().equals(localAddress.getAddress()) &&
                        m.getSocket().getLocalPort() == localAddress.getPort())
                  {
                     try
                     {
                        m.shutdownManager.incrementReferences();
                        return m;
                     }
                     catch (Exception e)
                     {
                        log.debug("manager shutting down: " + m);
                     }
                  }
               }
            }
         }
      }

      log.debug("There is no joinable MultiplexingManager. Creating new one.");
      m = new MultiplexingManager(socketFactory);
      m.bind(localAddress);
//      m.connect(remoteAddress, timeout);
      return m;
   }


/**
 * @param remoteAddress
 * @param localAddress
 * @param socketFactory
 * @return
 * @throws IOException
 */
   public static MultiplexingManager
   getAnExistingShareableManagerByAddressPair(InetSocketAddress remoteAddress,
                                              InetSocketAddress localAddress,
                                              SocketFactory socketFactory)
   throws IOException
   {
      log.debug("entering getaShareableManager(InetSocketAddress address)");
      MultiplexingManager m;

      // Check each of the shareable MultiplexingManagers connected to the target remote address, looking
      // for one which is not shutting down.
      synchronized(shareableMapLock)
      {
         HashSet managers = (HashSet) shareableManagers.get(remoteAddress);

         if (managers != null && !managers.isEmpty())
         {
            Iterator it = managers.iterator();

            while (it.hasNext())
            {
               m = (MultiplexingManager) it.next();

               if (socketFactory == m.getSocketFactory())
               {
                  if (m.getSocket().getLocalAddress().equals(localAddress.getAddress()) &&
                        m.getSocket().getLocalPort() == localAddress.getPort())
                  {
                     try
                     {
                        m.shutdownManager.incrementReferences();
                        return m;
                     }
                     catch (Exception e)
                     {
                        log.debug("manager shutting down: " + m);
                     }
                  }
               }
            }
         }
      }

      return null;
   }


/**
 * @param address
 * @return
 * @throws IOException
 */
   public static boolean checkForShareableManager(InetSocketAddress address) throws IOException
   {
      return checkForShareableManager(address, null);
   }


/**
 * @param address
 * @param socketFactory
 * @return
 * @throws IOException
 */
   public static boolean checkForShareableManager(InetSocketAddress address, SocketFactory socketFactory)
   throws IOException
   {
      log.debug("entering checkForShareableManager(InetSocketAddress address)");

      // Check each if there is at least one shareable MultiplexingManagers connected to the target remote address.
      synchronized (shareableMapLock)
      {
         HashSet managers = (HashSet) shareableManagers.get(address);

         if (managers != null && !managers.isEmpty())
         {
            Iterator it = managers.iterator();
            while (it.hasNext())
            {
               MultiplexingManager m = (MultiplexingManager) it.next();
               if (socketFactory == m.getSocketFactory())
                  return true;
            }
         }

         return false;
      }
   }


/**
 * @param localAddress
 * @param remoteAddress
 * @return
 */
   public static boolean checkForManagerByAddressPair(InetSocketAddress localAddress,
                                                      InetSocketAddress remoteAddress)
   {
      return checkForManagerByAddressPair(localAddress, remoteAddress, null);
   }


/**
 * @param localAddress
 * @param remoteAddress
 * @param socketFactory
 * @return
 */
   public static boolean checkForManagerByAddressPair(InetSocketAddress localAddress,
                                                      InetSocketAddress remoteAddress,
                                                      SocketFactory socketFactory)
   {
      log.debug("entering checkForManagerByAddressPair()");

      // Check each of the MultiplexingManagers connected to the target remote address, looking
      // for one bound to the local address.
      synchronized(remoteAddressMapLock)
      {
         HashSet managers = (HashSet) managersByRemoteAddress.get(remoteAddress);

         if (managers != null && !managers.isEmpty())
         {
            Iterator it = managers.iterator();

            while (it.hasNext())
            {
               MultiplexingManager m = (MultiplexingManager) it.next();

               if (socketFactory == m.getSocketFactory())
                  if (m.localSocketAddress.equals(localAddress))
                     return true;
            }
         }
      }

      return false;
   }


/**
 * @param localAddress
 * @param remoteAddress
 * @return
 */
   public static boolean checkForShareableManagerByAddressPair(InetSocketAddress localAddress,
                                                               InetSocketAddress remoteAddress)
   {
      return checkForShareableManagerByAddressPair(localAddress, remoteAddress, null);
   }


/**
 * @param localAddress
 * @param remoteAddress
 * @return
 */
   public static boolean checkForShareableManagerByAddressPair(InetSocketAddress localAddress,
                                                               InetSocketAddress remoteAddress,
                                                               SocketFactory socketFactory)
   {
      log.debug("entering checkForShareableManagerByAddressPair()");

      // Check each of the shareable MultiplexingManagers connected to the target remote address, looking
      // for one bound to the local address.
      synchronized (shareableMapLock)
      {
         HashSet managers = (HashSet) shareableManagers.get(remoteAddress);

         if (managers != null && !managers.isEmpty())
         {
            Iterator it = managers.iterator();

            while (it.hasNext())
            {
               MultiplexingManager m = (MultiplexingManager) it.next();

               if (socketFactory == m.getSocketFactory())
                  if (m.localSocketAddress.equals(localAddress))
                     return true;
            }
         }
      }

      return false;
   }


/**
 *
 * @param address
 * @throws IOException
 */
   public synchronized void bind(InetSocketAddress address) throws IOException
   {
      if (bound)
         throw new IOException("socket is already bound");

      socket.bind(address);
//      registerByLocalAddress(address);
      bound = true;
   }

/**
 *
 * @param address
 * @param timeout TODO
 * @throws IOException
 */
   public synchronized void connect(InetSocketAddress address, int timeout) throws IOException
   {
      if (connected)
      {
         if (socket.getRemoteSocketAddress().equals(address))
            return;
         else
            throw new IOException("socket is already connected");
      }

      if (socket == null)
         socket = new Socket();

      log.debug("connecting to: " + address);
      socket.connect(address, timeout);
      connected = true;
      setup();
   }


/**
 *
 * @return
 */
   public synchronized MultiplexingDataInputStream registerServerSocket(ServerSocket serverSocket) throws IOException
   {
      if (this.serverSocket != null && this.serverSocket != serverSocket)
      {
         log.error("[" + id + "]: " + "attempt to register a second server socket");
         log.error("current server socket: " + this.serverSocket.toString());
         log.error("new server socket:     " + serverSocket.toString());
         throw new IOException("attempt to register a second server socket");
      }

//      // check if still accepting clients - throws IOException otherwise
//      shutdownManager.incrementReferences();

      log.debug(serverSocket.toString());

      this.serverSocket = serverSocket;
      return new MultiplexingDataInputStream(getAnInputStream(SocketId.SERVER_SOCKET_ID, null));
   }


/**
 *
 * @param serverSocket
 * @throws IOException
 */
   public synchronized void unRegisterServerSocket(ServerSocket serverSocket) throws IOException
   {
      if (this.serverSocket != serverSocket)
      {
         log.error("server socket attempting unregister but is not registered");
         throw new IOException("server socket is not registered");
      }

      log.debug("server socket unregistering");
      removeAnInputStream(SocketId.SERVER_SOCKET_ID);
      this.serverSocket = null;
      shutdownManager.decrementReferences();
   }


/**
 *
 * @param socket
 * @return
 * @throws IOException
 */
   public synchronized MultiplexingInputStream registerSocket(VirtualSocket socket) throws IOException
   {
//      // check if still accepting clients - throws IOException otherwise
//      shutdownManager.incrementReferences();

      SocketId localSocketId = socket.getLocalSocketId();
      VirtualSocket currentSocket = (VirtualSocket) socketMap.put(localSocketId, socket);

      if (currentSocket != null)
      {
         String errorMessage = "attempting to register socket on currently used port:"
                              + currentSocket.getLocalVirtualPort();
         log.error(errorMessage);
         throw new IOException(errorMessage);
      }

      log.debug("registering virtual socket on port: " + localSocketId.getPort());
      registeredSockets.add(socket.getLocalSocketId());
      return getAnInputStream(localSocketId, socket);
   }


/**
 *
 * @param socket
 * @throws IOException
 */
   public synchronized void unRegisterSocket(VirtualSocket socket) throws IOException
   {
      log.debug("entering unRegisterSocket()");
      SocketId localSocketId = socket.getLocalSocketId();

      if (localSocketId == null)
         return;

      VirtualSocket currentSocket = (VirtualSocket) socketMap.remove(localSocketId);

      if (currentSocket == null)
      {
         String errorMessage = "attempting to unregister unrecognized socket: " + socket.getLocalSocketId().getPort();
         log.error(errorMessage);
         throw new IOException(errorMessage);
      }

      log.debug("unregistering virtual socket on port: " + localSocketId.getPort());
      registeredSockets.remove(socket.getLocalSocketId());
      removeAnInputStream(localSocketId);
      shutdownManager.decrementReferences();
   }


 /**
  *  FIXME Comment this
  *
  *
  */
    public synchronized void registerRemoteServerSocket() throws IOException
    {
       if (remoteServerSocketRegistered)
       {
          log.error("duplicate remote server socket registration");
          throw new IOException("duplicate remote server socket registration");
       }
       else
       {
          remoteServerSocketRegistered = true;

          // Now that remote MultiplexingManager has a VirtualServerSocket, we
          // add it to set of managers eligible to accept new clients.
          registerShareable(remoteSocketAddress);

          synchronized (threadsWaitingForRemoteServerSocket)
          {
             threadsWaitingForRemoteServerSocket.notifyAll();
          }
       }
    }


/**
 *  FIXME Comment this
 *
 *
 */
   public synchronized void unRegisterRemoteServerSocket()
   {
      if (!remoteServerSocketRegistered)
         log.error("no remote server socket is registered");
      else
      {
         remoteServerSocketRegistered = false;
         unregisterShareable();
      }
   }


   public synchronized boolean isRemoteServerSocketRegistered()
   {
      return remoteServerSocketRegistered;
   }


   public boolean waitForRemoteServerSocketRegistered()
   {
      if (remoteServerSocketRegistered)
         return true;

      synchronized (threadsWaitingForRemoteServerSocket)
      {
         threadsWaitingForRemoteServerSocket.add(Thread.currentThread());

         while (!remoteServerSocketRegistered)
         {
            try
            {
               threadsWaitingForRemoteServerSocket.wait();
            }
            catch (InterruptedException e)
            {
               log.info("interrupted waiting for registration of remote server socket");

               if (shutdown)
               {
                  threadsWaitingForRemoteServerSocket.remove(Thread.currentThread());
                  return false;
               }
            }
         }
      }

      threadsWaitingForRemoteServerSocket.remove(Thread.currentThread());
      return true;
   }

/**
 *  FIXME Comment this
 *
 * @throws IOException
 */
   public void incrementReferences() throws IOException
   {
      shutdownManager.incrementReferences();
   }


/**
 *  FIXME Comment this
 *
 * @throws IOException
 */
   public void decrementReferences() throws IOException
   {
      shutdownManager.decrementReferences();
   }


/**
 *
 * @return
 */
   public Collection getAllOutputStreams()
   {
      return outputStreamMap.values();
   }


/**
 *
 * @param socketId
 * @param socket TODO
 * @return
 * @throws IOException
 */
   public MultiplexingInputStream getAnInputStream(SocketId socketId, VirtualSocket socket) throws IOException
   {
      log.debug("getAnInputStream(): " + socketId.getPort());
      MultiplexingInputStream mis = (MultiplexingInputStream) inputStreamMap.get(socketId);

      if (mis != null)
         return mis;

      GrowablePipedOutputStream pos = new GrowablePipedOutputStream();
      mis = new MultiplexingInputStream(pos, this, socket);
      outputStreamMap.put(socketId, pos);
      inputStreamMap.put(socketId, mis);
      return mis;
   }


/**
 *
 * @return
 */
   public OutputMultiplexor getOutputMultiplexor()
   {
      return outputMultiplexor;
   }


/**
 *
 * @param socketId
 * @return
 */
   public OutputStream getOutputStreamByLocalSocket(SocketId socketId)
   {
      return (OutputStream) outputStreamMap.get(socketId);
   }


/**
 *
 * @return
 */
   public Protocol getProtocol()
   {
      return protocol;
   }


/**
 *  FIXME Comment this
 *
 * @return
 */
   public synchronized ServerSocket getServerSocket()
   {
      return serverSocket;
   }


/**
 *
 * @return
 */
   public Socket getSocket()
   {
      return socket;
   }


/**
 *
 * @param socketId
 * @return
 */
   public VirtualSocket getSocketByLocalPort(SocketId socketId)
   {
      return (VirtualSocket) socketMap.get(socketId);
   }


/**
 * @return
 */
   public SocketFactory getSocketFactory()
   {
      return socketFactory;
   }


/**
 *  FIXME Comment this
 *
 * @return
 */
   public boolean isBound()
   {
      return bound;
   }


/**
 *  FIXME Comment this
 *
 * @return
 */
   public boolean isConnected()
   {
      return connected;
   }


/**
 *  FIXME Comment this
 *
 * @return
 */
   public synchronized boolean isServerSocketRegistered()
   {
      return serverSocket != null;
   }


/**
 *
 */
   public boolean isShutdown()
   {
      return shutdown;
   }


/**
 *
 * @param socketId
 * @return
 * TODO: isn't used
 */
   public synchronized boolean isSocketRegistered(SocketId socketId)
   {
      return registeredSockets.contains(socketId);
   }


/**
 *  FIXME Comment this
 *
 * @return
 */
   public boolean respondToShutdownRequest()
   {
      return shutdownManager.respondToShutdownRequest();
   }


/**
 * @param socketFactory
 */
   public void setSocketFactory(SocketFactory socketFactory)
   {

   }


/********************************************************************************************
 *                     	protected methods and classes
 ********************************************************************************************/

/**
 *  FIXME Comment this
 * 
 * @param address
 */
   protected void registerByLocalAddress(InetSocketAddress address)
   {
      synchronized (localAddressMapLock)
      {
         localSocketAddress = address;
         HashSet managers = (HashSet) managersByLocalAddress.get(address);
         
         if (managers == null)
         {
            managers = new HashSet();
            managersByLocalAddress.put(address, managers);
         }
         
         managers.add(this);
         
         // allow searching on any local address
         localWildCardAddress = new InetSocketAddress(address.getPort());
         managers = (HashSet) managersByLocalAddress.get(localWildCardAddress);
         
         if (managers == null)
         {
            managers = new HashSet();
            managersByLocalAddress.put(localWildCardAddress, managers);
         }
         
         managers.add(this);
      }
      
//      managersByLocalAddress.put(address, this);
//      
//      // allow searching on any local address
//      localWildCardAddress = new InetSocketAddress(address.getPort());
//      managersByLocalAddress.put(localWildCardAddress, this);
   }
   
   
/**
 *  FIXME Comment this
 * 
 *
 */
   protected void unregisterByLocalAddress()
   {  
      synchronized (localAddressMapLock)
      {
         HashSet managers = null;
         
         if (localSocketAddress != null)
         {
            managers = (HashSet) managersByLocalAddress.get(localSocketAddress);
            
            if (managers != null)
            {
               managers.remove(this);
               
               if (managers.isEmpty())
                  managersByLocalAddress.remove(localSocketAddress);
            }
         }
         
         if (localWildCardAddress != null)
         {
            managers = (HashSet) managersByLocalAddress.get(localWildCardAddress);
            
            if (managers != null)
            {
               managers.remove(this);
               
               if (managers.isEmpty())
                  managersByLocalAddress.remove(localWildCardAddress);
            }
         } 
      }
   }
 
   
/**
 * 
 * @param address
 */
   protected void registerByRemoteAddress(InetSocketAddress address)
   {
      remoteSocketAddress = address;
      
      synchronized (remoteAddressMapLock)
      {
         HashSet managers = (HashSet) managersByRemoteAddress.get(address);
         
         if (managers == null)
         {
            managers = new HashSet();
            managers.add(this);
            managersByRemoteAddress.put(address, managers);
         }
         else
            managers.add(this);
      }
   }
  
   
/**
 *  FIXME Comment this
 * 
 *
 */
   protected void unregisterByRemoteAddress()
   {
      if (remoteSocketAddress != null)
      {
         synchronized (remoteAddressMapLock)
         {
            HashSet managers = (HashSet) managersByRemoteAddress.get(remoteSocketAddress);
            
            if (managers != null)
            {
               managers.remove(this);
               
               if (managers.isEmpty())
                  managersByRemoteAddress.remove(remoteSocketAddress);
            }
         }
      }
   }
   
   
/**
 * 
 * @param address
 */
   protected void registerShareable(InetSocketAddress address)
   {
      synchronized (shareableMapLock)
      {
         HashSet managers = (HashSet) shareableManagers.get(address);
         
         if (managers == null)
         {
            managers = new HashSet();
            managers.add(this);
            shareableManagers.put(address, managers);
         }
         else
            managers.add(this);
      }
   }
   
      
/**
 *  FIXME Comment this
 * 
 *
 */
   protected void unregisterShareable()
   {
      if (remoteSocketAddress != null)
      {
         synchronized (shareableMapLock)
         {
            HashSet managers = (HashSet) shareableManagers.get(remoteSocketAddress);
            
            if (managers != null)
            {
               managers.remove(this);
               
               if (managers.isEmpty())
                  shareableManagers.remove(remoteSocketAddress);
            }
         }
      }
   }
   
   
/**
 *  FIXME Comment this
 * 
 * @param socketId
 */
   protected void removeAnInputStream(SocketId socketId)
   {
      log.debug("entering removeAnInputStream(): " + socketId.getPort());
      InputStream is = (InputStream) inputStreamMap.remove(socketId);
      OutputStream os = (OutputStream) outputStreamMap.remove(socketId);
      
      if (is != null)
      {
         try
         {
            is.close();
         }
         catch (Exception ignored)
         {
            log.error("error closing PipedInputStream (" + socket.getPort() + ")", ignored);
         }
      }
      
      if (os != null)
      {
         try
         {
            os.close();
         }
         catch (Exception ignored)
         {
            log.error("error closing PipedOutputStream (" + socket.getPort() + ")", ignored);
         }
      }
   }
   
   
/**
 * 
 *
 */
   protected synchronized void shutdown()
   {
      log.debug("entering shutdown()");
      ShutdownThread shutdownThread = new ShutdownThread();
      shutdownThread.setName(shutdownThread.getName() + ":shutdown");
      shutdownThread.start();
   }
   
   
/**
 *  FIXME Comment this
 * 
 * @param pendingAction
 */
   protected void addToPendingActions(PendingAction pendingAction)
   {
      synchronized (pendingActions)
      {
         pendingActions.add(pendingAction);
         pendingActions.notifyAll();
      }
   }
   
   
/**
 * The motivation behind this class is to prevent the following problem.  Suppose MultiplexingManager A
 * is connected to MultiplexingManager B, and A decides to shut down.  Suppose A shuts down before B knows A
 * is shutting down, and suppose a VirtualSocket C starts up, finds B, and attaches itself to B.  Then C 
 * will have "died a-borning," in the words of Tom Paxton.  We need a handshake protocol to ensure a
 * proper shutdown process.  In the following, let A be the local MultiplexingManager and let B be its
 * remote peer.
 * 
 * There are two forms of synchronization in ShutdownManager.  incrementReferences() and decrementReferences()
 * maintain the reference count to its MultiplexingManager, and of course the changes to variable
 * referenceCount have to be synchronized.  Parallel to incrementReferences() and decrementReferences() are
 * the pair of methods reseverManager() and unreserveManager(), which are similar but intended for holding a
 * MultiplexingManger for more temporary applications.  See, for example, getaManagerByRemoteAddress().
 * 
 * There is also a need for distributed synchronization. When decrementReferences() on A decrements the 
 * referenceCount to 0, it indicates to B the desire of A to shut down. Since all of the virtual sockets 
 * on A are connected to virtual sockets on B, normally the request would be honored.  However, if a new virtual
 * socket attaches itself to B, then it would have > 0 clients, and it would refuse the request to shut down.
 * In any case, then request is made through Protocol.requestManagerShutdown(), which results in a call to
 * ShutdownManager.respondToShutdownRequest() on B, which is synchronized since it reads the 
 * readyToShutdown variable, which is modified by decrementReferences().  Here lies the danger of
 * distributed deadlock.  If decrementReferences() on A and B both start executing at about the same time, they
 * would both be waiting for a response from respondToShutdownRequest().  But each respondToShutdownRequest()
 * would be locked out because each decrementReferences() is blocked on i/o.  
 * 
 * The solution is to put the i/o operation in a separate thread, ShutdownRequestThread, and have decrementReferences()
 * enter a wait state, allowing respondToShutdownRequest() to execute.  So, on each end respondToShutdownRequest()
 * will return an answer ("go ahead and shut down", in particular), and each ShutdownRequestThread.run() will wake up
 * the waiting decrementReferences(), which will then run to completion.
 * 
 * Note also that while decrementReferences() is waiting, incrementReferences() can run.  However, before it waits,
 * decrementReferences() sets the shutdownRequestInProgress flag, and if incrementReferences() finds the flag set, it
 * will also enter a wait state and will take no action until the outstanding shutdown request is accepted or
 * rejected. 
 * 
 * Another issue is what to do if MultiplexingManager B responds negatively to A's request to shut down, not
 * because it has a new client but simply because some of its virtual sockets just haven't gotten around to 
 * closing yet.  When B's referenceCount finally goes to 0, it will send a shutdown request to A, and if A's
 * referenceCount is still 0, B will shut down.  But what about B?  If decrementReferences() gets a negative
 * response, it will start up a ShutdownMonitorThread, which, as long as readyToShutdown remains true, will
 * periodically check to see if remoteShutdown has been set to true.  If it has, ShutdownMonitorThread 
 * initiates the shut down of its enclosing MultiplexingManager.
 * 
 * reserveManager() interacts with decrementReferences() by preventing decrementReferences() from committing to
 * shutting down.  If reserveManager() runs first, it sets the flag reserved to true, which causes 
 * decrementReferences() to return without checking for referenceCount == 0.  If decrementReferences() runs 
 * first and finds referenceCount == 0 and gets a positve response from the remote manager, then reserveManager()
 * will throw an IOException.  But if decrementReferences() gets a negative response, it will start up a 
 * ShutdownMonitorThread, which runs as long as the flag readyToShutdown is true.  But reserveManager() will
 * set readyToShutdown to false, ending the ShutdownMonitorThread.  When unreserveManager() eventually runs
 * and sees referenceCount == 0, it increment referenceCount and call decrementReferences(), allowing the
 * shutdown attempt to proceed anew.  Note that when incrementReferences() runs successfully, it sets the flag
 * reserved to false, since incrementing referenceCount will also keep the MultiplexingManager alive.
 */
   protected class ShutdownManager
   {
      /** referenceCount keeps track of the number of clients of ShutdownManager's enclosing 
       *  MultiplexingManager. */
      private int referenceCount = 1;
      
      /** reserved is set to true to prevent the manager from shutting down without incrementing
       *  the reference count. */
      private boolean reserved = false;
      
      /** shutdownRequestInProgress remains true while a remote shutdown request is pending */
      private boolean shutdownRequestInProgress = false;
      
      /** readyToShutdown is set to true as long as long as referenceCount == 0.  It is interpreted
       *  by respondToShutdownRequest() as the local manager's willingness to shut down. */
      private boolean readyToShutdown = false;
      
      /** shutdownMonitorThread holds a reference to the most recently created ShutdownMonitorThread. */
      ShutdownMonitorThread shutdownMonitorThread;
      
      /** shutdown is set to true when the irrevocable decision has been made to shut down.  Once it
       *  is set to true, it is never set to false. */
      private boolean shutdown = false;
      
      /** remoteShutdown is set to true when the remote manager makes a shutdown request and 
       *  respondToShutdownRequest(), its agent for the request, discovers that the local manager
       *  is also willing to shut down.  It represents the fact that the remote manager will respond
       *  by deciding to shut down. */
      private boolean remoteShutdown = false;
      
      
      private class ShutdownRequestThread extends Thread
      { 
         public ShutdownRequestThread()
         {
            shutdownRequestInProgress = true;
         }
         
         public void run()
         {
            try
            {
               shutdown = protocol.requestManagerShutdown();
            }
            catch (SocketTimeoutException e)
            {
               log.debug("socket timeout exception in manager shutdown request");
            }
            catch (Exception e)
            {
               log.debug("i/o exception in manager shutdown request", e);
            }
            
            log.debug("ShutdownRequestThread.run(): done");
            shutdownRequestInProgress = false;
            
            synchronized(ShutdownManager.this)
            {
               ShutdownManager.this.notifyAll();
            }
         }
      }
      
      
/**
 * It is possible,due to a race condition, for there to be multiple ShutdownMonitorThreads running.
 * In particular, suppose
 * <p>
 * <ol>
 *   <li> decrementReferences() starts a ShutdownMonitorThread T1,
 *   <li> reserverManager() runs and calls T1.terminate(),
 *   <li> unreserveManager() runs, sees referenceCount == 0, and calls decrementReferences(),
 *   <li> decrementReferences() creates a new ShutdownMonitorThread T2,
 *   <li> T1 leaves sleep(1000) finds remoteShutdown == true and calls shutdown(), and
 *   <li> T2 finds remoteShutdown == true.
 * </ol>
 * <p>
 * For this reason, we turn on shutdown before calling shutdown() in T1, so that T2 will
 * not call shutdown().
 * 
 */
      private class ShutdownMonitorThread extends Thread
      {
         boolean running = true;
         
         public void terminate()
         {
            running = false;
         }
         
         public void run()
         {
            log.debug(socket.toString() + ": entering ShutdownMonitorThread");
            
            while (running)
            {
               try
               {
                  sleep(1000);
               }
               catch (InterruptedException ignored) {}

               synchronized (ShutdownManager.this)
               {
                  if (readyToShutdown && remoteShutdown && !shutdown)
                  {
                     log.debug("ShutdownMonitorThread: found remoteShutdown == true");
                     shutdown = true;
                     shutdown();
                     ShutdownManager.this.notifyAll();
                     return;
                  }
               }
            }
         }
      }
      
      
/**
 *  FIXME Comment this
 * 
 * @throws IOException
 */
      public synchronized void reserveManager() throws IOException
      {
         log.debug(socket.toString() + referenceCount);
         
         // If decrementReferences() grabbed the lock first, set referenceCount to 0, and initiated a
         // remote shutdown request, wait until the answer comes back.
         while (shutdownRequestInProgress)
         {
            try
            {
               wait();
            }
            catch (InterruptedException e)
            {
               // shouldn't happen
               log.error("interruption in ShutdownRequestThread");
            }
         }
         
         // 1. shutdown is true if and only if (a) ShutdownRequestThread returned an indication that the
         //    remote manager is shutting down, or (b) the wait() in decrementReferences() for the return
         //    of ShutdownRequestThread timed out and decrementReferences() set shutdown to true;
         // 2. remoteShutdown is true if and only if the remote manager initiated a shutdown request
         //    which found that the local manager was ready to shut down. In this case, the remote 
         //    manager would go ahead and shut down, and the local manager will inevitably shutdown as well.
         
         if (shutdown || remoteShutdown)
            throw new IOException("manager shutting down");
         
         readyToShutdown = false;
         reserved = true;
         
         if (shutdownMonitorThread != null)
            shutdownMonitorThread.terminate();

         // wake up decrementReferences() if it is waiting
         notifyAll();
      }
      
      
/**
 *  FIXME Comment this
 * 
 *
 */
      public synchronized void unreserveManager()
      {
         log.debug(socket.toString() + referenceCount);
         
         if (!reserved)
         {
            log.error("attempting to unreserve a MultiplexingManager that was not reserved: " + socket.toString());
            return;
         }
         
         reserved = false;

         // If referenceCount == 0, it is because it was decremented by decrementReferences(). But if 
         // reserveManager() was able to run, it was because decrementReferences() did not succeed in
         // negotiating a shutdown.  Either it found reserved == true and gave up, or it received a
         // negative reply from the remote manager and started up a ShutdownMonitorThread, which would
         // have terminated when reserveManager() set readyToShutdown to false.  It is therefore
         // appropriate to give decrementReferences() another opportunity to shut down.
         if (referenceCount == 0)
         {
            referenceCount++;
            decrementReferences();
         }
      }
/**
 *
 */
      public synchronized void incrementReferences() throws IOException
      {
         log.debug(socket.toString() + referenceCount);
         
         // If decrementReferences() grabbed the lock first, set referenceCount to 0, and initiated a
         // remote shutdown request, wait until the answer comes back.
         while (shutdownRequestInProgress)
         {
            try
            {
               wait();
            }
            catch (InterruptedException e)
            {
               // shouldn't happen
               log.error("interruption in ShutdownRequestThread");
            }
         }
         
         // 1. shutdown is true if and only if (a) ShutdownRequestThread returned an indication that the
         //    remote manager is shutting down, or (b) the wait() in decrementReferences() for the return
         //    of ShutdownRequestThread timed out and decrementReferences() set shutdown to true;
         // 2. remoteShutdown is true if and only if the remote manager initiated a shutdown request
         //    which found that the local manager was ready to shut down. In this case, the remote 
         //    manager would go ahead and shut down, and the local manager will inevitably shutdown as well.
         
         if (shutdown || remoteShutdown)
            throw new IOException("not accepting new clients");
         
         readyToShutdown = false;
         reserved = false;
         referenceCount++;
         log.debug(socket.toString() + referenceCount);
                  
         if (shutdownMonitorThread != null)
            shutdownMonitorThread.terminate();

         // wake up decrementReferences() if it is waiting
         notifyAll();
      }
      
      
/**
 *  FIXME Comment this
 * 
 *
 */
      public synchronized void decrementReferences()
      {
         referenceCount--;
         log.debug(socket.toString() + referenceCount);
         
         if (reserved)
         {
            log.debug(socket.toString() + ": reserved == true");
            return;
         }
         
         if (referenceCount == 0)
         {
            readyToShutdown = true;
            
            if (isConnected())
            {
               ShutdownRequestThread shutdownRequestThread = new ShutdownRequestThread();
               shutdownRequestThread.setName(shutdownRequestThread.getName() + ":shutdownRequest");
               shutdownRequestThread.setDaemon(true);
               log.debug("starting ShutdownRequestThread: " + shutdownRequestThread.toString());
               shutdownRequestThread.start();
               
               try
               {
                  wait(5000);
               }
               catch (InterruptedException e)
               {
                  // shouldn't happen
                  log.error("interrupt in ShutdownRequestThread");
               }

               log.debug(socket.toString() + shutdown);
               log.debug(socket.toString() + shutdownRequestThread.isAlive());
               
               // If shutdownRequestInProgress is still true, we assume that the peer MultiplexingManager
               // has shut down or is inaccessible, and we shut down.
               if (shutdownRequestInProgress)
               {
                  shutdown = true;
                  
                  // turn off shutdownRequestInProgress in case incrementReferences() or reserveManager()
                  // are waiting
                  shutdownRequestInProgress = false;
               }
            }
            else // !isConnected()
               shutdown = true;
            
            
            if (shutdown)
            {
               shutdown();
               
               // wake up incrementReferences() if it is waiting
               notifyAll();
            }
            else
            {
               shutdownMonitorThread = new ShutdownMonitorThread();
               shutdownMonitorThread.setName(shutdownMonitorThread.getName() + ":shutdownMonitor");
               shutdownMonitorThread.start();
            }
         }
      }
      
      
/**
 *  FIXME Comment this
 * 
 * @return
 */
      protected synchronized boolean respondToShutdownRequest()
      {
         log.debug(socket.toString() + readyToShutdown);
         log.debug(socket.toString() + shutdown);
         
         if (readyToShutdown)
         {
            remoteShutdown = true;
            log.debug(socket.toString() + ": respondToShutdownRequest(): set remoteShutdown to true");
         }
         
         return readyToShutdown;
      }
      
      
/**
 *  FIXME Comment this
 * 
 * @return
 */
      protected boolean isShutdown()
      {
         return shutdown;
      }
   }
   
   
/**
 * 
 */
   protected class ShutdownThread extends Thread
   
   {
      public void run()
      {
         log.debug(socket.toString() + ": manager shutting down");  
         
         // Unregister this MultiplexingManager by local address(es)
         unregisterByLocalAddress();
         
         // Unregister this MultiplexingManager by remote address
         unregisterByRemoteAddress();
            
         // Remove this MultiplexingManager from set of shareable managers
         unregisterShareable();
         
         // When we get here, we don't know how far setting up progressed.
         if (outputThread != null)
         {
            // give output thread a chance to flush pending messages before shutting down socket
            outputThread.shutdown();
            
            try
            {  
               outputThread.join();
               log.debug("manager: joined output thread");
            }
            catch (InterruptedException ignored)
            {
               log.debug("interrupt exception waiting for write thread");
            }  
         }
         
         if (socket != null)
         {
            try 
            {
               socket.close();
               log.debug("manager: closed socket");
            }
            catch (IOException e)
            {
               log.error("manager: unable to close socket");
            }
         }
         
         if (backChannelThread != null)
         {
            backChannelThread.shutdown();      
            
            try
            {   
               backChannelThread.join();
               log.debug("manager: joined back channel thread");
            }
            catch (InterruptedException ignored)
            {
               log.debug("manager: interrupted exception waiting for back channel thread");
            }
         }
         
         if (inputThread != null)
         {
            inputThread.shutdown();
            
            try
            {  
               inputThread.join();
               log.debug("manager: joined input thread");
            }
            catch (InterruptedException ignored)
            {
               log.debug("manager: interrupted exception waiting for read thread");
            }    
         }
         
         removeAnInputStream(SocketId.PROTOCOL_SOCKET_ID);
         removeAnInputStream(SocketId.SERVER_SOCKET_ID);
         removeAnInputStream(SocketId.SERVER_SOCKET_CONNECT_ID);
         removeAnInputStream(SocketId.SERVER_SOCKET_VERIFY_ID);
         removeAnInputStream(SocketId.BACKCHANNEL_SOCKET_ID);
         
         if (pendingActionThread != null)
         {
            pendingActionThread.shutdown();
            
            try
            {
               pendingActionThread.join();
               log.debug("manager: joined socketAndStreamClose thread");
            }
            catch (InterruptedException e)
            {
               log.debug("manager: interrupted exception waiting for socket close thread to shut down");
            }
         }
         
         shutdown = true;
         log.debug("manager shut down");
      }
   }

   
   protected class PendingActionThread extends StoppableThread
   {
      private List pendingActionsTemp = new ArrayList();
      
      protected void doInit()
      {
         log.debug("PendingActionThread starting");
      }

      protected void doRun()
      {  
         synchronized (pendingActions)
         {
            while (pendingActions.isEmpty())
            {
               try
               {
                  pendingActions.wait();
               }
               catch (InterruptedException ignored)
               {
                  if (!isRunning())
                     return;
               }
            }

            pendingActionsTemp.addAll(pendingActions);
            pendingActions.clear();
         }
         
         Iterator it = pendingActionsTemp.iterator();
         
         while (it.hasNext())
         {
            Object o = it.next();
            
//            if (o instanceof VirtualSocket)
//            {
//               VirtualSocket socket = (VirtualSocket) o;
//               socket.doClose();
//            }
            if (o instanceof PendingAction)
               ((PendingAction) o).doAction();
            else
               log.error("object in closePendingSockets has invalid type: " + o.getClass());
         }
         
         pendingActionsTemp.clear();
      }
      
      public void shutdown()
      {
         log.debug("entering close thread shutdown()");
         super.shutdown();
         interrupt();
      }

      protected void doShutDown()
      {
         log.debug("PendingActionThread shutting down");
      }
   }
}