/*
* JBoss, Home of Professional Open Source
* Copyright 2005, JBoss Inc., and individual contributors as indicated
* by the @authors tag. See the copyright.txt in the distribution for a
* full listing of individual contributors.
*
* This is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2.1 of
* the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free
* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
* 02110-1301 USA, or see the FSF site: http://www.fsf.org.
*/

/*
 * Created on Jul 22, 2005
 */
 
package org.jboss.remoting.transport.multiplex;


import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.net.Socket;
import java.net.SocketException;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;

import org.jboss.logging.Logger;
import org.jboss.remoting.transport.multiplex.utility.StoppableThread;

/**
 * <p>
 * Copyright (c) 2005
 * <p>
 * @author <a href="mailto:r.sigal@computer.org">Ron Sigal</a>
 *
 * TODO: size limited payloads
 */
public class OutputMultiplexor
{
   protected static final Logger log = Logger.getLogger(OutputMultiplexor.class);
   protected static final int MARKER = 255;
   protected static final int MAX_CHUNK_LENGTH = 2048;
   static final int BRACKETS_ALL = -1;
   static final int BRACKETS_NONE = -2;

   /** The queue of messages to be processed by the output thread */
   private List writeQueue = Collections.synchronizedList(new LinkedList());
   private SocketId previousDestinationId = SocketId.INITIAL_OUTPUT_SOCKET_ID;

   private MultiplexingManager manager;
   private OutputStream os;
   private Message pendingMessage;
   private byte[] outputBytes = new byte[4096];
   private boolean tracing;



/**
 * @param socket TODO
 *
 */
   protected OutputMultiplexor(MultiplexingManager manager, Socket socket) throws IOException
   {
      this.manager = manager;
//      os = new BufferedOutputStream(socket.getOutputStream(), 2048);
      os = socket.getOutputStream();
      tracing = log.isTraceEnabled();
   }


   public OutputThread getAnOutputThread()
   {
      return new OutputThread();
   }


   public void write(SocketId socketId, byte[] content) throws IOException
   {
      write(socketId, content, BRACKETS_NONE);
   }
   
   
/**
 *
 * @param socketId
 * @param content
 * @throws InterruptedException
 */
   protected void write(SocketId socketId, byte[] content, int brackets) throws IOException
   {
      if (tracing)
      {
         String messageEnd = "";
         
         if (content.length > 0)
            messageEnd = ": [" + (0xff & content[0]) + "]";
         
         log.trace("OutputMultiplexor.write(): queueing "
               + content.length + " bytes for socket " + socketId.getPort() + messageEnd);
      }

      synchronized (writeQueue)
      {
         if (!writeQueue.isEmpty())
         {
            Message message = (Message) writeQueue.get(writeQueue.size() - 1);
            
            if (message.getDestination().equals(socketId) && message.hasCompatibleBrackets(brackets))
            {
               message.addContent(content);
               return;
            }
         }
         
         writeQueue.add(new Message(socketId, content, brackets));
         writeQueue.notifyAll();
      }

      if (tracing)
      {
         String messageEnd = "";
         
         if (content.length > 0)
            messageEnd = ": [" + (0xff & content[0]) + "]";
         
         log.trace("OutputMultiplexor.write(): queued "
               + content.length + " bytes for socket " + socketId.getPort() + messageEnd);
      }
   }


/**
*
*/
   class OutputThread extends StoppableThread
   {

      private final Logger log = Logger.getLogger(OutputMultiplexor.OutputThread.class);

      private int dataOutCount = 0;
      private boolean socketIsOpen = true;
      private boolean threadTracing;


      public OutputThread()
      {
         threadTracing = log.isTraceEnabled();
      }


      /**
       *
       */
      public void shutdown()
      {
         super.shutdown();
         interrupt();
      }


      protected void doInit()
      {
         log.debug("output thread starting");
      }


      protected void doRun()
      {
         while (socketIsOpen && (!writeQueue.isEmpty() || isRunning()))
         {
            try
            {
               synchronized(writeQueue)
               {
                  if (writeQueue.isEmpty())
                     writeQueue.wait();

                  pendingMessage = (Message) writeQueue.remove(0);
               }

               SocketId remoteSocketId =  pendingMessage.getDestination();
               int start = pendingMessage.getStart();
               int length = Math.min(pendingMessage.getLength(), MAX_CHUNK_LENGTH);
               encode(remoteSocketId, pendingMessage.getContent(), start, length, os);
               
               if (threadTracing)
                  log.trace("output thread wrote: " + length + " bytes to socket " + remoteSocketId.getPort());
               
               dataOutCount += length;

               if (length < pendingMessage.getLength())
               {
                  pendingMessage.markUsed(length);
                  
                  synchronized (writeQueue)
                  {
                     if (!writeQueue.isEmpty())
                     {
                        ListIterator it = writeQueue.listIterator();
                        boolean processed = false;
                        int remotePort = remoteSocketId.getPort();
                        int brackets = pendingMessage.getBrackets();
                        
                        while (it.hasNext())
                        {
                          Message message = (Message) it.next(); 
                          
                          if (message.brackets(remotePort))
                          {
                             it.previous();
                             it.add(pendingMessage);
                             processed = true;
                             break;
                          }
                          
                          if (message.getDestination().equals(remoteSocketId) 
                              && (BRACKETS_NONE == message.getBrackets() || brackets == message.getBrackets()))
                          {
                             pendingMessage.addContent(message.getContent(), message.getStart(), message.getLength());
                             it.set(pendingMessage);
                             processed = true;
                             break;
                          }
                        }
                        
                        if (!processed)
                        {
                           writeQueue.add(pendingMessage);
                        }
                     }
                     else
                     {
                        writeQueue.add(pendingMessage);
                     }
                  }
               }

               pendingMessage = null;

               if (interrupted()) // outside of writeQueue.take()
                  throw new InterruptedException();
            }
            catch (InterruptedException e)
            {
               handleError("output thread: interrupted", e);
            }
            catch (SocketException e)
            {
               handleError("output thread: socket exception", e);
            }
            catch (IOException e)
            {
               handleError("output thread: i/o error", e);
            }
         }

         log.debug("output thread: socketIsConnected: " + socketIsOpen);
         log.debug("output thread: writeQueue.isEmpty(): " + writeQueue.isEmpty());
         log.debug("output thread: running: " + running);
         log.debug("output thread: pendingMessage ==  " + pendingMessage);
      }


      /**
       *
       */
      protected void doShutDown()
      {
         //           try
         //           {
         //              os.close();
         //           }
         //           catch (Exception ignored)
         //           {
         //              if (trace)
         //                 log.trace(ignored.getMessage(), ignored);
         //           }

         //    TODO: work out
         //           if (terminatedOnError)
         //              multiplexor.shutdown();

         log.debug("output thread: data bytes out: " + dataOutCount);
         log.debug("output thread shutting down");

         try
         {
            while (pendingMessage != null)
            {
               log.debug("waiting for encode() to write final message");
               Thread.sleep(1000);
            }
         }
         catch (InterruptedException ignored)
         {
            log.debug(ignored);
         }
      }


   /**
    *
    * @param bytes
    * @param start TODO
    * @param length TODO
    * @param os
    * @throws IOException
    */
      protected void encode(SocketId destinationId, byte[] bytes,  int start, int length, OutputStream os) throws IOException
      {
         if (threadTracing)
         {
            String messageEnd = "";
            
            if (length > 0)
               messageEnd = " [" + bytes[start] + "]: " + destinationId.getPort();
            
            log.trace("encode(): writing " + length + " bytes" + messageEnd);
         }
         
         int position = 0;
         
         if (!destinationId.equals(previousDestinationId))
         {
            if (threadTracing)
            {
               log.trace("encode(): writing: MARKER");
               byte[] byteArray = destinationId.toByteArray();
               for (int i = 0; i < byteArray.length; i++)
                  log.trace("encode(): writing: " + byteArray[i]);
               log.trace("encode(): writing: MARKER");
            }
            
            previousDestinationId = destinationId;
            outputBytes[0] = (byte) MARKER;
            int len = destinationId.toByteArray().length;
            System.arraycopy(destinationId.toByteArray(), 0, outputBytes, 1, len);
            position = len + 1;
            outputBytes[position++] = (byte) MARKER;
         }
         
         if (threadTracing)
            log.trace("encode(): writing " + length + " bytes to socketId " + destinationId.getPort() );
         
         for (int i = start; i < start + length; i++)
         {
            if (threadTracing)
               log.info("" + bytes[i]);
            
            outputBytes[position++] = bytes[i];
            
            if (bytes[i] == (byte) MARKER)
               outputBytes[position++] = (byte) MARKER;
         }
         
         os.write(outputBytes, 0, position);
         os.flush();

         if (threadTracing)
         {
            String messageEnd = "";
            
            if (length > 0)
               messageEnd = ": [" + bytes[start] + "]";
           
            log.trace("encode(): wrote " + length + " bytes" + messageEnd);         
         }
      }



      /**
       *
       * @param message
       * @param e
       */
      protected void handleError(String message, Throwable e)
      {
         if (e instanceof SocketException)
         {
            socketIsOpen = false;
            super.shutdown();
            return;
         }
         
         if (log != null)
         {
            if (e instanceof InterruptedException || e instanceof IOException)
            {
               if (tracing)
                  log.trace(message, e);
            }
            else
               log.debug(message, e);
         }

         //    TODO
         //           running = false;
         //           terminatedOnError = true;
      }
   }



/**
 *
 */
   private static class Message
   {
      private SocketId socketId;
      private ByteArrayOutputStream baos = new ByteArrayOutputStream();
      private int start;
      private int length;
      private int brackets;

      public Message(SocketId socketId, byte[] content, int brackets) throws IOException
      {
         this.socketId = socketId;
         this.brackets = brackets;
         baos.write(content);
         length = content.length;
      }


      public SocketId getDestination()
      {
         return socketId;
      }

      public byte[] getContent()
      {
         return baos.toByteArray();
      }
      
      public void addContent(byte[] bytes) throws IOException
      {
         baos.write(bytes);
         length += bytes.length;
      }
      
      public void addContent(byte[] bytes, int start, int length)
      {
         baos.write(bytes, start, length);
         this.length += length;
      }
      
      public int getStart()
      {
         return start;
      }
      
      public int getLength()
      {
         return length;
      }
      
      public int getBrackets()
      {
         return brackets;
      }
      
      public void markUsed(int used)
      {
         length -= used;
         
         if (length <= 0)
         {
            start = 0;
            length = 0;
            baos.reset();
         }
         else
         {
            start += used;
         }
      }
      
      public boolean brackets(int b)
      {
         if (brackets == BRACKETS_ALL)
            return true;
         
         if (brackets == BRACKETS_NONE)
            return false;
         
         return (brackets == b);
      }
      
      public boolean hasCompatibleBrackets(int b)
      {
         if (brackets == BRACKETS_ALL || b == BRACKETS_NONE)
            return true;
         
         return (brackets == b);
      }
   }

}
