/*
* JBoss, Home of Professional Open Source
* Copyright 2005, JBoss Inc., and individual contributors as indicated
* by the @authors tag. See the copyright.txt in the distribution for a
* full listing of individual contributors.
*
* This is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2.1 of
* the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free
* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
* 02110-1301 USA, or see the FSF site: http://www.fsf.org.
*/

/*
 * Created on Jul 23, 2005
 */
 
package org.jboss.remoting.transport.multiplex;


import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.InterruptedIOException;
import java.io.OutputStream;
import java.net.Socket;
import java.net.SocketTimeoutException;

import org.jboss.logging.Logger;
import org.jboss.remoting.transport.multiplex.utility.StoppableThread;

/**
 * 
 * <p>
 * Copyright (c) 2005
 * <p>
 * @author <a href="mailto:r.sigal@computer.org">Ron Sigal</a>
 */
public class Protocol
{
   protected static final Logger log = Logger.getLogger(Protocol.class);
   
   // message types
   public static final int MP_CONNECT 	 	  		   = 0;
   public static final int MP_CONNECTED 	  		   = 1;
   public static final int MP_VERIFY_CONNECTION        = 2;
//   public static final int MP_INPUT_SHUTDOWN 		   = 3;
   public static final int MP_OUTPUT_SHUTDOWN 		   = 4;
   public static final int MP_DISCONNECT	  		   = 5;
   public static final int MP_REGISTER_REMOTE_SERVER   = 6;
   public static final int MP_UNREGISTER_REMOTE_SERVER = 7;
   public static final int MP_REQUEST_MANAGER_SHUTDOWN = 8;
   public static final int MP_ERROR 	 	  		   = 9;
   public static final int MP_TRUE					   = 10;
   public static final int MP_FALSE					   = 11;
   
   private MultiplexingManager manager;
   
   /** underlying MultiplexingInputStream used by connect() to set timeout */
   private MultiplexingInputStream protocolMultiplexingInputStream;
   
   /** InputStream used to receive synchronous messages */
   private DataInputStream protocolInputStream;
   
   /** OutputStream used to send synchronous messages */
   private DataOutputStream protocolOutputStream;
   
   /** OutputStream used by connect() to communicate with a ServerSocket */
   private DataOutputStream serverSocketOutputStream;
   
   /** InputStream used by acceptConnect() to communicate with a ServerSocket */
   private DataInputStream serverSocketInputStream;

   /** InputStream for receiving asynchronous remote messages */
   private DataInputStream backChannelInputStream;
  
   /** OutputStream for sending messages to remote backChannelInputStream */
   private MultiplexingOutputStream backChannelOutputStream;
   
   
/**
 * 
 * @param manager
 * @throws IOException
 */
   public Protocol(MultiplexingManager manager) throws IOException
   {
      this.manager = manager;
      Socket socket = manager.getSocket();
      
      protocolMultiplexingInputStream = manager.getAnInputStream(SocketId.PROTOCOL_SOCKET_ID, null);
      protocolMultiplexingInputStream.setTimeout(10000);
      protocolInputStream = new DataInputStream(protocolMultiplexingInputStream);
      protocolOutputStream = new DataOutputStream(new MultiplexingOutputStream(manager, SocketId.PROTOCOL_SOCKET_ID));
      serverSocketOutputStream = new DataOutputStream(new MultiplexingOutputStream(manager, SocketId.SERVER_SOCKET_ID));
      backChannelOutputStream = new MultiplexingOutputStream(manager, SocketId.BACKCHANNEL_SOCKET_ID); 
      backChannelInputStream = new DataInputStream(manager.getAnInputStream(SocketId.BACKCHANNEL_SOCKET_ID, null));;
   }
           
   
/**
 *  FIXME Comment this
 * @param is TODO
 * @param socketId
 * @return
 * @throws IOException
 */
   public SocketId connect(MultiplexingDataInputStream is, SocketId socketId) throws IOException
   {
      return connect(is, socketId, 0);
   }
   
   
   /**
    *  FIXME Comment this
    * @param is TODO
    * @param socketId
    * 
    * @return
    * @throws IOException
    */
      public SocketId connect(MultiplexingDataInputStream is, SocketId socketId, int timeout) throws IOException
      {
         log.debug("entering Protocol.connect()");
         long start = System.currentTimeMillis();
         int timeLeft = 0;
         int savedTimeout = is.getTimeout();
         
         synchronized (serverSocketOutputStream)
         {
            serverSocketOutputStream.write(MP_CONNECT);
            log.debug("Protocol.connect(): wrote: CONNECT (" + MP_CONNECT + ")");
            serverSocketOutputStream.writeInt(socketId.getPort());
            log.debug("Protocol.connect(): wrote port: " + socketId.getPort());
         }
         
         try
         {
            if (timeout > 0)
               if ((timeLeft = timeout - (int)(System.currentTimeMillis() - start)) <= 0)
                  throw new SocketTimeoutException();
               
            is.setTimeout(timeLeft);
            int messageType = is.read();
            log.debug("Protocol.connect(): read message type: " + messageType);
            
            switch (messageType)
            {
               case MP_CONNECTED:
                  if (timeout > 0)
                     if ((timeLeft = timeout - (int) (System.currentTimeMillis() - start)) <= 0)
                        throw new SocketTimeoutException("connect timed out");
                     
                  is.setTimeout(timeLeft);
                  int remotePort = is.readInt();
                  log.debug("Protocol.connect(): read port: " + remotePort);
                  return new SocketId(remotePort);
                  
               default:
                  log.error("Protocol.connect(): expecting a CONNECTED message: received: " + messageType);
                  throw new IOException("Protocol.connect(): expecting a CONNECTED message: received: " + messageType);
            }
         }
         catch (SocketTimeoutException e)
         {
            log.info("timeout in Protocol.connect()");
            
//            // The normal response to a connect request is 5 bytes.  Even though they were not all received
//            // before the timeout, they will eventually be sent, assuming that all is well with the network
//            // and the remote MultiplexingManager.  Since we are giving up on this connect attempt,
//            // we want those 5 bytes to be skipped.  If there is a problem preventing all 5 bytes from being
//            // sent, then the InputStream is useless anyway.
//            is.setSkip(5);
            throw e;
         }
         catch (Exception e)
         {
            log.error(e);
            e.printStackTrace();
            StackTraceElement[] stes = e.getStackTrace();
            for (int i = 0; i < stes.length; i++)
               log.error(stes[i].toString());
            throw new IOException(e.getMessage());
         }
         finally
         {
            is.setTimeout(savedTimeout);
         }
      }
      
   
/**
 *  FIXME Comment this
 * @param is
 * @param timeout TODO
 * 
 * @return
 * @throws IOException
 */
   public SocketId acceptConnect(MultiplexingDataInputStream is, int timeout) throws IOException
   {
      log.debug("entered acceptConnect()");
      long start = System.currentTimeMillis();
      int timeLeft = timeout;
      int savedTimeout = is.getTimeout();
      
      try
      {
         is.setTimeout(timeLeft);
         int messageType = is.read();
         log.debug("Protocol.acceptConnect(): read message type: " + messageType);
         
         switch (messageType)
         {
            case MP_CONNECT:
               if (timeout > 0)
                  if ((timeLeft = timeout - (int)(System.currentTimeMillis() - start)) <= 0)
                     throw new SocketTimeoutException();
                  
               is.setTimeout(timeLeft);
               int remotePort = is.readInt();
               log.debug("Protocol.acceptConnect(): read port: " + remotePort);
               return new SocketId(remotePort);
               
            case -1:
               log.info("Protocol.acceptConnect(): end of file");
               throw new IOException("end of file");
               
            default:
               log.error("Protocol.acceptConnect: expecting a CONNECT message: received: " + messageType);
               throw new IOException("Protocol.acceptConnect: expecting a CONNECT message: received: " + messageType);
         }
      }
      catch (SocketTimeoutException e)
      {
         log.info("timeout in Protocol.acceptConnect()");
         
//         // The normal connect request is 5 bytes.  Even though they were not all received
//         // before the timeout, they will eventually be sent, assuming that all is well with the network
//         // and the remote MultiplexingManager.  Since we are giving up on this connect attempt,
//         // we want those 5 bytes to be skipped.  If there is a problem preventing all 5 bytes from being
//         // sent, then the InputStream is useless anyway.
//         is.setSkip(5);
         throw e;
      }
      finally
      {
         is.setTimeout(savedTimeout);
      }
   }
   
   
/**
 *  FIXME Comment this
 * @param os TODO
 * @param port
 * 
 * @throws IOException
 */
   public void answerConnect(DataOutputStream os, int port) throws IOException
   {
      os.write(MP_CONNECTED);
      log.debug("Protocol.answerConnect(): wrote: CONNECTED (" + MP_CONNECTED + ")");
      os.writeInt(port);
      log.debug("Protocol.answerConnect(): wrote port: " + port);
   }
   

/**
 *  FIXME Comment this
 * @param os TODO
 * 
 * @throws IOException
 */
   public void verifyConnect(OutputStream os) throws IOException
   {
      try
      {
         os.write(MP_VERIFY_CONNECTION);
         log.debug("verifyConnect(): wrote: VERIFY_CONNECTION (" + MP_VERIFY_CONNECTION + ")");
      }
      catch (IOException e)
      {
         log.error("i/o error in verifyConnect()");
         throw e;
      }
   }
   
   
/**
 *  FIXME Comment this
 * @param is TODO
 * @param timeout TODO
 * 
 *
 */
   public void getConnectVerification(MultiplexingDataInputStream is, int timeout) throws IOException
   {
      log.debug("entered getConnectVerification()");
      int savedTimeout = is.getTimeout();
      
      try
      {
         is.setTimeout(timeout);
         is.read();
         log.debug("getConnectVerification(): received response");
      }
      catch (SocketTimeoutException e)
      {
         log.info("timeout in Protocol.getConnectVerification()");
         
//         // The normal verification response is 1 byte.  Even if it were not received
//         // before the timeout, it will eventually be sent, assuming that all is well with the network
//         // and the remote MultiplexingManager.  Since we are giving up on this connect attempt,
//         // we want that byte to be skipped.  If there is a problem preventing it from being
//         // sent, then the InputStream is useless anyway.
//         is.setSkip(1);
         throw e;
      }
      catch (IOException e)
      {
         log.error("i/o error in getConnectVerification()");
      }
      finally
      {
         is.setTimeout(savedTimeout);
      }
   }

///**
// *  FIXME Comment this
// * 
// * @param socketId
// */
//   public synchronized void notifyInputShutdown(SocketId socketId)
//   {
//      int port = socketId.getPort();
//      
//      try
//      {
//         backChannelOutputStream.write(MP_INPUT_SHUTDOWN);
//         backChannelOutputStream.writeInt(port);
//         log.info("Protocol.notifyInputShutdown(): wrote: INPUT_SHUTDOWN (" + MP_INPUT_SHUTDOWN + ") for port: " + port);         
//      }
//      catch (IOException ignored)
//      {
//         log.error("Protocol.notifyInputShutdown(): unable to send MP_INPUT_SHUTDOWN message to port: " + port);
//      }
//   }
   
   
/**
 *  FIXME Comment this
 * 
 * @param socketId
 */
   public void notifyOutputShutdown(SocketId socketId)
   {
      int port = socketId.getPort();
      
      try
      {
         synchronized (backChannelOutputStream)
         {
            backChannelOutputStream.write(MP_OUTPUT_SHUTDOWN, port);
            backChannelOutputStream.writeInt(port, port);
         }
         
         log.debug("Protocol.notifyOutputShutdown(): wrote: OUTPUT_SHUTDOWN (" + MP_OUTPUT_SHUTDOWN + ") for port: " + port);         
      }
      catch (IOException ignored)
      {
         log.error("Protocol.notifyOutputShutdown(): unable to send MP_OUTPUT_SHUTDOWN message to port: " + port);
      }
   }
   
   
/**
 *  FIXME Comment this
 * 
 * @param socketId
 */
   public void disconnect(SocketId socketId)
   {
      int port = socketId.getPort();
      
      try
      {
         synchronized (backChannelOutputStream)
         {
            backChannelOutputStream.write(MP_DISCONNECT, port);
            backChannelOutputStream.writeInt(port, port);
         }
         
         log.debug("Protocol.disconnect(): wrote: DISCONNECT (" + MP_CONNECTED + ") for port: " + port);         
      }
      catch (IOException ignored)
      {
         log.error("Protocol.disconnect(): unable to send DISCONNECT message to port: " + port);
      }
   }
   
   
/**
 *  FIXME Comment this
 * @param timeout 
 * 
 *
 */
   public void registerRemoteServerSocket(int timeout) throws IOException
   {
      int answer = MP_FALSE; 
      
      synchronized (protocolInputStream)
      {
         synchronized (backChannelOutputStream)
         {
            backChannelOutputStream.write(MP_REGISTER_REMOTE_SERVER);
         }
         
         log.debug("Protocol.registerRemoteServerSocket(): wrote: REGISTER_REMOTE_SERVER (" + MP_REGISTER_REMOTE_SERVER + ")");         
         int savedTimeout = protocolMultiplexingInputStream.getTimeout();
         protocolMultiplexingInputStream.setTimeout(timeout);
         
         try
         {
            answer = protocolInputStream.read();
         }
         finally
         {
            protocolMultiplexingInputStream.setTimeout(savedTimeout); 
         }
      }
      
      log.debug("Protocol.registerRemoteServerSocket(): read: " + (answer == MP_TRUE ? "true" : "false"));
      
      if (answer == MP_FALSE)
         throw new IOException("unable to register remote socket");
   }
   

/**
 *  FIXME Comment this
 * 
 *
 */
   public void unregisterRemoteServerSocket()
   {
      try
      {
         synchronized (backChannelOutputStream)
         {
            backChannelOutputStream.write(MP_UNREGISTER_REMOTE_SERVER);
         }
         
         log.debug("Protocol.disconnect(): wrote: UNREGISTER_REMOTE_SERVER (" + MP_UNREGISTER_REMOTE_SERVER + ")");         
      }
      catch (IOException ignored)
      {
         log.error("Protocol.unregisterRemoteServerSocket(): unable to send UNREGISTER_REMOTE_SERVER");
      }
   }
   
   
   public boolean requestManagerShutdown() throws IOException
   {
      int b;
      
      synchronized (protocolInputStream)
      {
         synchronized (backChannelOutputStream)
         {
            backChannelOutputStream.write(MP_REQUEST_MANAGER_SHUTDOWN, OutputMultiplexor.BRACKETS_ALL);
         }
         
         log.debug("Protocol.requestManagerShutdown(): wrote: REQUEST_MANAGER_SHUTDOWN (" + MP_REQUEST_MANAGER_SHUTDOWN + ")");
         b = protocolInputStream.read();
      }
      
      boolean answer = (b == MP_TRUE) ? true : false;
      log.debug("Protocol.requestManagerShutdown(): read: " + answer);
      return answer;
   }
   
   
/**
 *  FIXME Comment this
 * 
 * @return
 */
   public BackChannelThread getBackChannelThread()
   {
      return new BackChannelThread(manager, backChannelInputStream);
   }
   
   
/**
 * 
 */
   class BackChannelThread extends StoppableThread
   {
      
      private MultiplexingManager manager;
      private DataInputStream is;
      
      
      public BackChannelThread(MultiplexingManager manager, DataInputStream is)
      {
         this.manager = manager;
         this.is = is;
      }

      
/**
 * 
 */
      public void shutdown()
      {
         log.debug("back channel thread: beginning shut down");
         super.shutdown();
         interrupt();
      }
      
      
/**
 * 
 */
      protected void doInit()
      {
         log.debug("back channel thread starting");
      }

      
/**
 * 
 */
      protected void doRun()
      {   
         int messageType;
         int port;
         VirtualSocket socket;
         int answer;
         
         try
         {
            messageType = is.read();
            log.debug("back channel thread: read message type: " + messageType);
            
            switch (messageType)
            {      
//               	case MP_INPUT_SHUTDOWN:
//               	   
//            	   port = is.readInt();
//            	   log.info("back channel thread: read INPUT_SHUTDOWN for port: " + port);
//            	   socket = manager.getSocketByLocalPort(new SocketId(port)); 
//            	   
//            	   if (socket == null)
//            	   {
//            	      log.info("back channel thread (INPUT_SHUTDOWN): unable to retrieve socket at port: " + port);
//            	   }
//            	   else
//            	   {
//            	      socket.handleRemoteInputShutDown();
//            	   }
//            	   
//            	   break;
            	   
               	case MP_OUTPUT_SHUTDOWN:
               	   
               	   port = is.readInt();
            	   log.debug("back channel thread: read OUTPUT_SHUTDOWN for port: " + port);
            	   socket = manager.getSocketByLocalPort(new SocketId(port)); 
            	   
            	   if (socket == null)
            	   {
            	      log.info("back channel thread (OUTPUT_SHUTDOWN): unable to retrieve socket at port: " + port);
            	   }
            	   else
            	   {
            	      socket.handleRemoteOutputShutDown();
            	   }
            	   
            	   break;
               	   
            	case MP_DISCONNECT:
            	   
            	   port = is.readInt();
            	   log.debug("back channel thread: read DISCONNECT for port: " + port);
            	   socket = manager.getSocketByLocalPort(new SocketId(port)); 
            	   
            	   if (socket == null)
            	   {
            	      log.info("back channel thread (DISCONNECT): unable to retrieve socket at port: " + port);
            	   }
            	   else
            	   {
            	      socket.handleRemoteDisconnect();
            	   }
            	   
            	   break;
               	   
            	case MP_REGISTER_REMOTE_SERVER:
            	   
            	   // remote VirtualServerSocket is starting up
              	   log.debug("back channel thread: read REGISTER_REMOTE_SERVER");
              	   answer = MP_TRUE;
              	   
              	   try
            	   {
              	      manager.registerRemoteServerSocket();
            	   }
              	   catch (Exception e)
              	   {
              	      answer = MP_FALSE;
              	      log.info("back channel thread: unable to register remote server");
              	   }
              	   
               	   protocolOutputStream.write(answer);
            	   break;
            	   
            	case MP_UNREGISTER_REMOTE_SERVER:
            	   // remote VirtualServerSocket is shutting down
              	   log.debug("back channel thread: read UNREGISTER_REMOTE_SERVER");
            	   manager.unRegisterRemoteServerSocket();
            	   break;
            	   
            	case MP_REQUEST_MANAGER_SHUTDOWN:
            	   // remote MultiplexingManager is shutting down
              	   log.debug("back channel thread: read REQUEST_MANAGER_SHUTDOWN");
              	   answer = manager.respondToShutdownRequest() ? MP_TRUE : MP_FALSE;
              	   log.debug("back channel thread: writing " + answer);
            	   protocolOutputStream.write(answer);
            	   break;
            	   
            	default:
            	   log.error("unexpected message type in back channel thread: " + messageType);
            }
         }
         catch (InterruptedIOException e)
         {
            if (isRunning())
               log.error("back channel thread: i/o interruption", e);
            else
               log.debug("back channel thread: i/o interruption");
         }
         catch (IOException e)
         { 
            if (isRunning())
            {
               log.error("back channel thread: i/o error", e);
               running = false;
            }
            else
               log.debug("back channel thread: i/o error");
         }
      }

      
/**
 * 
 */
      protected void doShutDown()
      {
         log.debug("back channel thread shutting down");
      }
      
   }
}
