/*
* JBoss, Home of Professional Open Source
* Copyright 2005, JBoss Inc., and individual contributors as indicated
* by the @authors tag. See the copyright.txt in the distribution for a
* full listing of individual contributors.
*
* This is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2.1 of
* the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free
* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
* 02110-1301 USA, or see the FSF site: http://www.fsf.org.
*/

package org.jboss.remoting.transport.servlet.web;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import javax.management.MBeanServer;
import javax.management.MBeanServerInvocationHandler;
import javax.management.MalformedObjectNameException;
import javax.management.ObjectName;
import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.ServletInputStream;
import javax.servlet.ServletOutputStream;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import org.jboss.logging.Logger;
import org.jboss.mx.util.MBeanServerLocator;
import org.jboss.remoting.transport.servlet.ServletServerInvokerMBean;

/**
 * The servlet that receives the inital http request for the ServletServerInvoker.
 *
 * @author <a href="mailto:tom.elrod@jboss.com">Tom Elrod</a>
 */
public class ServerInvokerServlet extends HttpServlet
{
   private static Logger log = Logger.getLogger(ServerInvokerServlet.class);
   private ServletServerInvokerMBean servletInvoker;

   /**
    * Initializes the servlet.
    */
   public void init(ServletConfig config) throws ServletException
   {
      super.init(config);
      ObjectName localInvokerName = null;
      try
      {
         // See if the servlet is bound to a particular invoker
         String name = config.getInitParameter("invokerName");
         if(name != null)
         {
            localInvokerName = new ObjectName(name);
            log.debug("localInvokerName=" + localInvokerName);
         }
      }
      catch(MalformedObjectNameException e)
      {
         throw new ServletException("Failed to build invokerName", e);
      }

      // Lookup the MBeanServer
      MBeanServer mbeanServer = MBeanServerLocator.locateJBoss();
      if(mbeanServer == null)
      {
         throw new ServletException("Failed to locate the MBeanServer");
      }
      else
      {
         servletInvoker = (ServletServerInvokerMBean)
               MBeanServerInvocationHandler.newProxyInstance(mbeanServer,
                                                             localInvokerName,
                                                             ServletServerInvokerMBean.class,
                                                             false);

      }
   }

   /**
    * Destroys the servlet.
    */
   public void destroy()
   {

   }

   /**
    * Read a MarshalledInvocation and dispatch it to the target JMX object
    * invoke(Invocation) object.
    *
    * @param request  servlet request
    * @param response servlet response
    */
   protected void processRequest(HttpServletRequest request, HttpServletResponse response)
         throws ServletException, IOException
   {
      boolean trace = log.isTraceEnabled();
      if(trace)
      {
         log.trace("processRequest, ContentLength: " + request.getContentLength());
         log.trace("processRequest, ContentType: " + request.getContentType());
      }

      int bufferSize = 1024;
      byte[] byteBuffer = new byte[bufferSize];
      ByteArrayOutputStream byteOutputStream = new ByteArrayOutputStream();

      int pointer = 0;
      int contentLength = request.getContentLength();
      ServletInputStream inputStream = request.getInputStream();
      int amtRead = inputStream.read(byteBuffer);

      while(amtRead > 0)
      {
         byteOutputStream.write(byteBuffer, pointer, amtRead);
         //pointer+=amtRead;
         if(amtRead < bufferSize && byteOutputStream.size() >= contentLength)
         {
            //done reading, so process
            break;
         }
         amtRead = inputStream.read(byteBuffer);
      }
      byteOutputStream.flush();
      byte[] totalByteArray = byteOutputStream.toByteArray();

      byte[] out = servletInvoker.processRequest(request, totalByteArray, response);
      ServletOutputStream outStream = response.getOutputStream();
      outStream.write(out);
      outStream.flush();
      outStream.close();
      //response.setContentLength(out.length);
   }

   /**
    * Handles the HTTP <code>GET</code> method.
    *
    * @param request  servlet request
    * @param response servlet response
    */
   protected void doGet(HttpServletRequest request, HttpServletResponse response)
         throws ServletException, IOException
   {
      processRequest(request, response);
   }

   /**
    * Handles the HTTP <code>POST</code> method.
    *
    * @param request  servlet request
    * @param response servlet response
    */
   protected void doPost(HttpServletRequest request, HttpServletResponse response)
         throws ServletException, IOException
   {
      processRequest(request, response);
   }

   /**
    * Returns a short description of the servlet.
    */
   public String getServletInfo()
   {
      return "Servlet front to JBossRemoting servlet server invoker.";
   }

}