/*
* JBoss, Home of Professional Open Source
* Copyright 2005, JBoss Inc., and individual contributors as indicated
* by the @authors tag. See the copyright.txt in the distribution for a
* full listing of individual contributors.
*
* This is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2.1 of
* the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free
* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
* 02110-1301 USA, or see the FSF site: http://www.fsf.org.
*/
package org.jboss.remoting.util.socket;

import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.security.KeyManagementException;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.NoSuchProviderException;
import java.security.Provider;
import java.security.SecureRandom;
import java.security.UnrecoverableKeyException;
import java.security.cert.CertificateException;
import java.util.Map;
import java.net.Socket;
import java.net.UnknownHostException;
import java.net.InetAddress;
import javax.net.SocketFactory;
import javax.net.ssl.KeyManager;
import javax.net.ssl.SSLContext;
import javax.net.ssl.SSLSocket;
import javax.net.ssl.TrustManager;
import javax.net.ssl.KeyManagerFactory;
import javax.net.ssl.X509KeyManager;
import javax.net.ssl.TrustManagerFactory;
import javax.net.ssl.SSLSocketFactory;
import javax.net.ssl.SSLSessionContext;
import javax.net.ssl.SSLSession;

/**
 * @author <a href="mailto:tom.elrod@jboss.com">Tom Elrod</a>
 */
public class RemotingSSLSocketFactory extends SSLSocketFactory
{
   private SSLContext sslContext = null;
   private Provider provider = null;
   private String providerName = null;

   private String protocol = "TLS";  // default protocol

   private KeyManager[] keyManagers = null;
   private TrustManager[] trustManagers = null;
   private SecureRandom secureRandom = null;

   private SSLSocketFactory socketFactory = null;

   private String keyStoreFilePath = null;
   private String keyStorePassword = null;
   private String keyStoreType = null;
   private String keyAlias = null;
   private String algorithm = null;

   private String trustAlgorithm = null;
   private String trustStoreType = null;
   private String trustStorePassword = null;
   private String trustStoreFilePath = null;

   private Map config = null;
   
   private boolean useClientMode = true;

   public static final String REMOTING_KEY_STORE_FILE_PATH = "org.jboss.remoting.keyStore";
   public static final String STANDARD_KEY_STORE_FILE_PATH = "javax.net.ssl.keyStore";
   public static final String REMOTING_KEY_STORE_PASSWORD = "org.jboss.remoting.keyStorePassword";
   public static final String STANDARD_KEY_STORE_PASSWORD = "javax.net.ssl.keyStorePassword";
   /**
    * Default key store type if one not set as bean property, via config, or via system property.
    * Value is 'JKS'.
    */
   public static final String DEFAULT_KEY_STORE_TYPE = "JKS";
   public static final String REMOTING_KEY_STORE_TYPE = "org.jboss.remoting.keyStoreType";
   public static final String STANDARD_KEY_STORE_TYPE = "javax.net.ssl.keyStoreType";

   /**
    * Default algorithm if one net set as bean property or via config.
    * Value is 'SunX509'.
    */
   public static final String DEFAULT_ALGORITHM = "SunX509";
   public static final String REMOTING_ALGORITHM = "org.jboss.remoting.algorithm";

   public static final String REMOTING_KEY_ALIAS = "org.jboss.remoting.keyAlias";

   public static final String REMOTING_TRUST_ALGORITHM = "org.jboss.remoting.truststoreAlgorithm";
   public static final String REMOTING_TRUST_STORE_TYPE = "org.jboss.remoting.trustStoreType";
   public static final String STANDARD_TRUST_STORE_TYPE = "javax.net.ssl.trustStoreType";
   public static final String REMOTING_TRUST_STORE_PASSWORD = "org.jboss.remoting.trustStorePassword";
   public static final String STANDARD_TRUST_STORE_PASSWORD = "javax.net.ssl.trustStorePassword";
   public static final String REMOTING_TRUST_STORE_FILE_PATH = "org.jboss.remoting.trustStore";
   public static final String STANDARD_TRUST_STORE_FILE_PATH = "javax.net.ssl.trustStore";
   
   public static final String REMOTING_USE_CLIENT_MODE = "org.jboss.remoting.useClientMode";

   public RemotingSSLSocketFactory(Map config)
   {
      this.config = config;
   }

   public void setProtocol(String protocol)
   {
      if(protocol != null && protocol.length() > 0)
      {
         this.protocol = protocol;
      }
      else
      {
         throw new IllegalArgumentException("Can not set remoting socket factory with null protocol");
      }
   }

   public String getProtocol()
   {
      return this.protocol;
   }

   public void setProvider(Provider provider)
   {
      this.provider = provider;
   }

   public void setProviderName(String providerName)
   {
      this.providerName = providerName;
   }

   public Provider getProvider()
   {
      if(provider != null)
      {
         return provider;
      }
      else if(sslContext != null)
      {
         return sslContext.getProvider();
      }
      else
      {
         return null;
      }
   }

   private void init() throws IOException
   {
      try
      {
         if(provider != null)
         {
            sslContext = SSLContext.getInstance(protocol, provider);
         }
         else if(providerName != null)
         {
            sslContext = SSLContext.getInstance(protocol, providerName);
         }
         else
         {
            sslContext = SSLContext.getInstance(protocol);
         }

         // load all managers
         keyManagers = loadKeyManagers();
         trustManagers = loadTrustManagers();
         secureRandom = getSecureRandom();

         sslContext.init(keyManagers, trustManagers, secureRandom);

         socketFactory = sslContext.getSocketFactory();

//      SSLSessionContext sslSessionConext = sslContext.getClientSessionContext();
//      SSLSession session = sslSessionConext.getSession(???);
      }
      catch(Exception e)
      {
         IOException ioe = new IOException("Error creating Remoting SSL Socket Factory.");
         ioe.setStackTrace(e.getStackTrace());
         throw ioe;
      }
   }

   public void setSecureRandom(SecureRandom secureRandom)
   {
      this.secureRandom = secureRandom;
   }

   private SecureRandom getSecureRandom()
   {
      if(secureRandom != null)
      {
         return secureRandom;
      }
      else
      {
         secureRandom = new SecureRandom();
      }
      return secureRandom;
   }

   private TrustManager[] loadTrustManagers()
         throws NoSuchProviderException, NoSuchAlgorithmException, IOException, CertificateException, KeyStoreException
   {
      TrustManager[] trustManagers = null;

      String tsType = getTrustStoreType();
      String tsPasswd = getTrustStorePassword();
      String tsPath = getTrustStoreFile();

      KeyStore trustStore = loadKeyStore(tsType, tsPath, tsPasswd);
      if(trustStore != null)
      {
         String tsAlg = getTrustAlgorithm();
         TrustManagerFactory trustMgrFactory = null;

         if(provider != null)
         {
            trustMgrFactory = TrustManagerFactory.getInstance(tsAlg, provider);
         }
         else if(providerName != null)
         {
            trustMgrFactory = TrustManagerFactory.getInstance(tsAlg, providerName);
         }
         else
         {
            trustMgrFactory = TrustManagerFactory.getInstance(tsAlg);
         }
         trustMgrFactory.init(trustStore);
         trustManagers = trustMgrFactory.getTrustManagers();
      }
     return trustManagers;

   }

   private KeyManager[] loadKeyManagers()
         throws NoSuchProviderException, NoSuchAlgorithmException, IOException, CertificateException,
                KeyStoreException, UnrecoverableKeyException
   {
      KeyManager[] keyManagers = null;

      String ksPath = getKeyStoreFile();
      String ksPasswd = getKeyStorePassword();
      String ksType = getKeyStoreType();
      if(ksPath != null)
      {
         KeyStore keyStore = loadKeyStore(ksType, ksPath, ksPasswd);

         if(keyStore != null)
         {
            String alg = getAlgorithm();
            String alias = getKeyAlias();

            // check that keystore contains supplied alias (if there is one)
            if(alias != null)
            {
               boolean containsAlias = keyStore.isKeyEntry(alias);
               if(!containsAlias)
               {
                  // can not continue as supplied alias does not exist as key entry
                  throw new IOException("Can not find key entry for key store (" + ksPath + ") with given alias (" + alias + ")");
               }
            }
            KeyManagerFactory keyMgrFactory = null;

            if(provider != null)
            {
               keyMgrFactory = KeyManagerFactory.getInstance(alg, provider);
            }
            else if(providerName != null)
            {
               keyMgrFactory = KeyManagerFactory.getInstance(alg, providerName);
            }
            else
            {
               keyMgrFactory = KeyManagerFactory.getInstance(alg);
            }
            keyMgrFactory.init(keyStore, ksPasswd.toCharArray());

            keyManagers = keyMgrFactory.getKeyManagers();

            // if alias provided, use helper impl to hard wire alias name to be used
            if(alias != null)
            {
               //TODO: -TME Need careful review of if this is really needed or not.
               for(int x = 0; x < keyManagers.length; x++)
               {
                  keyManagers[x] = new RemotingKeyManager((X509KeyManager)keyManagers[x], alias);
               }
            }

         }
      }
      return keyManagers;
   }

   private KeyStore loadKeyStore(String keyStoreType, String keyStorePath, String keyStorePassword)
         throws KeyStoreException, NoSuchProviderException, IOException, NoSuchAlgorithmException, CertificateException
   {
      KeyStore keyStore = null;

      if(provider != null)
      {
         keyStore = KeyStore.getInstance(keyStoreType, provider);
      }
      else if(providerName != null)
      {
         keyStore = KeyStore.getInstance(keyStoreType, providerName);
      }
      else
      {
         keyStore = KeyStore.getInstance(keyStoreType);
      }

      // now that keystore instance created, need to load data from file
      InputStream keyStoreInputStream = null;

      try
      {
         // first try to get stream using thread's classloader
         keyStoreInputStream = Thread.currentThread().getContextClassLoader().getResourceAsStream(keyStorePath);

         if(keyStoreInputStream == null) // couldn't find by relative path, so will try absolute
         {
            keyStoreInputStream = new FileInputStream(keyStorePath);
         }

         // is ok for password to be null, as will just be used to check integrity of store
         char[] password = keyStorePassword != null ? keyStorePassword.toCharArray() : null;
         keyStore.load(keyStoreInputStream, password);
      }
      finally
      {
         if(keyStoreInputStream != null)
         {
            try
            {
               keyStoreInputStream.close();
            }
            catch(IOException e)
            {
               // no op
            }
            keyStoreInputStream = null;
         }
      }

      return keyStore;
   }

   public void setTrustStoreFile(String trustStoreFilePath)
   {
      this.trustStoreFilePath = trustStoreFilePath;
   }

   private String getTrustStoreFile()
   {
      if(trustStoreFilePath != null)
      {
         return trustStoreFilePath;
      }
      else
      {
         if(config != null)
         {
            String path = (String)config.get(REMOTING_TRUST_STORE_FILE_PATH);
            if(path != null && path.length() > 0)
            {
               trustStoreFilePath = path;
            }
         }
         if(trustStoreFilePath == null)
         {
            trustStoreFilePath = System.getProperty(STANDARD_TRUST_STORE_FILE_PATH);
         }
      }
      return trustStoreFilePath;
   }

   private String getKeyStoreFile()
   {
      if(keyStoreFilePath != null)
      {
         return keyStoreFilePath;
      }
      else
      {
         if(config != null)
         {
            String path = (String) config.get(REMOTING_KEY_STORE_FILE_PATH);
            if(path != null && path.length() > 0)
            {
               keyStoreFilePath = path;
            }
         }
         if(keyStoreFilePath == null)
         {
            keyStoreFilePath = System.getProperty(STANDARD_KEY_STORE_FILE_PATH);
         }
      }
      return keyStoreFilePath;
   }

   public void setKeyStoreFile(String keyStoreFilePath)
   {
      this.keyStoreFilePath = keyStoreFilePath;
   }

   public void setTrustStorePassword(String trustStorePassword)
   {
      this.trustStorePassword = trustStorePassword;
   }

   private String getTrustStorePassword()
   {
      if(trustStorePassword != null)
      {
         return trustStorePassword;
      }
      else
      {
         if(config != null)
         {
            String passwd = (String)config.get(REMOTING_TRUST_STORE_PASSWORD);
            if(passwd != null && passwd.length() > 0)
            {
               trustStorePassword = passwd;
            }
         }
         if(trustStorePassword == null)
         {
            trustStorePassword = System.getProperty(STANDARD_TRUST_STORE_PASSWORD);
         }
         if(trustStorePassword == null)
         {
            trustStorePassword = getKeyStorePassword();
         }
      }
      return trustStorePassword;
   }

   private String getKeyStorePassword()
   {
      if(keyStorePassword != null)
      {
         return keyStorePassword;
      }
      else
      {
         if(config != null)
         {
            String passwd = (String)config.get(REMOTING_KEY_STORE_PASSWORD);
            if(passwd != null && passwd.length() > 0)
            {
               keyStorePassword = passwd;
            }
         }
         if(keyStorePassword == null)
         {
            keyStorePassword = System.getProperty(STANDARD_KEY_STORE_PASSWORD);
         }
      }
      return keyStorePassword;
   }
   
   private void initializeUseClientMode()
   {
      if (config.get(REMOTING_USE_CLIENT_MODE).equals("false"))
         useClientMode = false;
   }

   public void setKeyStorePassword(String keyStorePassword)
   {
      this.keyStorePassword = keyStorePassword;
   }

   public void setTrustStoreType(String trustStoreType)
   {
      this.trustStoreType = trustStoreType;
   }

   public String getTrustStoreType()
   {
      if(trustStoreType != null)
      {
         return trustStoreType;
      }
      else
      {
         if(config != null)
         {
            String type = (String)config.get(REMOTING_TRUST_STORE_TYPE);
            if(type != null && type.length() > 0)
            {
               trustStoreType = type;
            }
         }
         if(trustStoreType == null)
         {
            trustStoreType = System.getProperty(STANDARD_TRUST_STORE_TYPE);
         }
         if(trustStoreType == null)
         {
            trustStoreType = DEFAULT_KEY_STORE_TYPE;
         }
      }
      return trustStoreType;
   }

   public String getKeyStoreType()
   {
      if(keyStoreType != null)
      {
         return keyStoreType;
      }
      else
      {
         if(config != null)
         {
            String type = (String)config.get(REMOTING_KEY_STORE_TYPE);
            if(type != null && type.length() > 0)
            {
               keyStoreType = type;
            }
         }
         if(keyStoreType == null)
         {
            keyStoreType = System.getProperty(STANDARD_KEY_STORE_TYPE);
         }
         if(keyStoreType == null)
         {
            keyStoreType = DEFAULT_KEY_STORE_TYPE;
         }
      }
      return keyStoreType;
   }

   public void setKeyStoreType(String keyStoreType)
   {
      this.keyStoreType = keyStoreType;
   }

   public void setTrustAlgorithm(String algorithm)
   {
      this.trustAlgorithm = algorithm;
   }

   public String getTrustAlgorithm()
   {
      if(trustAlgorithm != null)
      {
         return trustAlgorithm;
      }
      else
      {
         if(config != null)
         {
            String alg = (String)config.get(REMOTING_TRUST_ALGORITHM);
            if(alg != null && alg.length() > 0)
            {
               trustAlgorithm = alg;
            }
         }
         if(trustAlgorithm == null)
         {
            trustAlgorithm = getAlgorithm();
         }
      }
      return trustAlgorithm;
   }

   public String getAlgorithm()
   {
      if(algorithm != null)
      {
         return algorithm;
      }
      else
      {
         if(config != null)
         {
            String alg = (String)config.get(REMOTING_ALGORITHM);
            if(alg != null && alg.length() > 0)
            {
               algorithm = alg;
            }
         }
         if(algorithm == null)
         {
            algorithm = DEFAULT_ALGORITHM;
         }
      }
      return algorithm;
   }

   public void setAlgorithm(String algorithm)
   {
      this.algorithm = algorithm;
   }

   public String getKeyAlias()
   {
      if(keyAlias != null)
      {
         return keyAlias;
      }
      if(config != null)
      {
         keyAlias = (String)config.get(REMOTING_KEY_ALIAS);
      }
      return keyAlias;
   }

   public void setKeyAlias(String alias)
   {
      this.keyAlias = alias;
   }

   public String[] getDefaultCipherSuites()
   {
      if(socketFactory == null)
      {
         try
         {
            init();
         }
         catch(IOException e)
         {
            throw new RuntimeException(e);
         }
      }
      return socketFactory.getDefaultCipherSuites();
   }

   public String[] getSupportedCipherSuites()
   {
      if(socketFactory == null)
      {
         try
         {
            init();
         }
         catch(IOException e)
         {
            throw new RuntimeException(e);
         }
      }
      return socketFactory.getSupportedCipherSuites();
   }

   public Socket createSocket() throws IOException
   {
      if(socketFactory == null)
      {
            init();
      }

      SSLSocket s = (SSLSocket) socketFactory.createSocket();
      
      if (!useClientMode)
         s.setUseClientMode(false);
      
      return socketFactory.createSocket();
   }
   
   public Socket createSocket(Socket socket, String string, int i, boolean b) throws IOException
   {
      if(socketFactory == null)
      {
            init();
      }

      SSLSocket s = (SSLSocket) socketFactory.createSocket(socket, string, i, b);
      
      if (!useClientMode)
         s.setUseClientMode(false);
      
      return s;
   }

   public Socket createSocket(String string, int i) throws IOException, UnknownHostException
   {
      if(socketFactory == null)
      {
            init();
      }

      SSLSocket s = (SSLSocket) socketFactory.createSocket(string, i);
      
      if (!useClientMode)
         s.setUseClientMode(false);
      
      return s;
   }

   public Socket createSocket(InetAddress inetAddress, int i) throws IOException
   {
      if(socketFactory == null)
      {
            init();
      }

      SSLSocket s = (SSLSocket) socketFactory.createSocket(inetAddress, i);
      
      if (!useClientMode)
         s.setUseClientMode(false);
      
      return s;
   }

   public Socket createSocket(String string, int i, InetAddress inetAddress, int i1) throws IOException, UnknownHostException
   {
      if(socketFactory == null)
      {
            init();
      }

      SSLSocket s = (SSLSocket) socketFactory.createSocket(string, i, inetAddress, i1);
      
      if (!useClientMode)
         s.setUseClientMode(false);
      
      return s;
   }

   public Socket createSocket(InetAddress inetAddress, int i, InetAddress inetAddress1, int i1) throws IOException
   {
      if(socketFactory == null)
      {
            init();
      }

      SSLSocket s = (SSLSocket) socketFactory.createSocket(inetAddress, i, inetAddress1, i1);
      
      if (!useClientMode)
         s.setUseClientMode(false);
      
      return s;
   }
}